from __future__ import annotations

from typing import Protocol, runtime_checkable

from pydantic import BaseModel


class TextPart(BaseModel):
    text: str


class InlineBlobPart(BaseModel):
    encode_type: str = "base64"
    mime_type: str
    data: str


type Part = TextPart | InlineBlobPart


class Content(BaseModel):
    parts: list[Part]

    @property
    def text(self) -> str:
        content_text = ""
        for part in self.parts:
            if isinstance(part, TextPart):
                content_text += part.text
        return content_text

    @staticmethod
    def from_text(text: str) -> Content:
        return Content(parts=[TextPart(text=text)])


type ContentUnion = str | Content


@runtime_checkable
class Promptable(Protocol):
    def to_prompt(self) -> ContentUnion: ...



from enum import StrEnum
from typing import Literal

from litellm.types.utils import ChatCompletionMessageToolCall, Usage
from pydantic import BaseModel

from judy_core.messages import Task
from judy_core.types import ContentUnion


class OutputEventKind(StrEnum):
    TOOL_CALL = "tool_call"
    TASK_FAIL = "task_fail"
    TASK_SUCCESS = "task_success"
    CREATE_AGENT_FOLDER = "create_agent_folder"
    DEBUG_MODEL_INPUT = "debug_llm_input"
    DEBUG_MODEL_OUTPUT = "debug_llm_output"
    DEBUG_TOOL_RESULT = "debug_tool_result"
    DEBUG_TASK_LOG = "debug_task_log"


class ToolCall(BaseModel):
    kind: Literal[OutputEventKind.TOOL_CALL] = OutputEventKind.TOOL_CALL
    source: str
    tool_call_id: str
    name: str
    arguments: str


class ToolResult(BaseModel):
    id: str
    name: str
    content: ContentUnion


class TaskFailResult(BaseModel):
    kind: Literal[OutputEventKind.TASK_FAIL] = OutputEventKind.TASK_FAIL
    source: str = ""
    content: str


class TaskSuccessResult(BaseModel):
    kind: Literal[OutputEventKind.TASK_SUCCESS] = OutputEventKind.TASK_SUCCESS
    content: str


class CreateAgentFolder(BaseModel):
    kind: Literal[OutputEventKind.CREATE_AGENT_FOLDER] = (
        OutputEventKind.CREATE_AGENT_FOLDER
    )
    folder_name: str


class DebugModelInput(BaseModel):
    kind: Literal[OutputEventKind.DEBUG_MODEL_INPUT] = OutputEventKind.DEBUG_MODEL_INPUT
    source: str
    role: str
    content: str


class DebugModelOutput(BaseModel):
    kind: Literal[OutputEventKind.DEBUG_MODEL_OUTPUT] = (
        OutputEventKind.DEBUG_MODEL_OUTPUT
    )
    source: str
    response_id: str
    model: str | None
    tool_calls: list[ChatCompletionMessageToolCall] | None
    content: str | None
    reasoning_content: str | None
    usage: Usage | None
    hidden_params: dict | None
    cost_money: float | None
    cost_time: float


class DebugToolResult(BaseModel):
    kind: Literal[OutputEventKind.DEBUG_TOOL_RESULT] = OutputEventKind.DEBUG_TOOL_RESULT
    source: str
    tool_call_id: str
    name: str
    content: ContentUnion
    time_cost: float


class DebugTaskLog(BaseModel):
    kind: Literal[OutputEventKind.DEBUG_TASK_LOG] = OutputEventKind.DEBUG_TASK_LOG
    level: str = "warning"
    content: str


type InputEvent = Task | ToolResult
type OutputEvent = (
    ToolCall
    | TaskFailResult
    | TaskSuccessResult
    | CreateAgentFolder
    | DebugModelInput
    | DebugModelOutput
    | DebugToolResult
    | DebugTaskLog
)



from dataclasses import dataclass
from typing import List, Optional


def middle_truncate(text: str, limit: int, newline_marker: bool) -> tuple[str, bool]:
    if len(text) <= limit:
        return text, False

    head_len = limit // 2
    tail_len = limit - head_len
    suffix = text[-tail_len:]
    marker = f"...[TRUNCATED {len(text) - limit} chars]..."
    if newline_marker:
        marker = f"\n{marker}\n"
    return text[:head_len] + marker + suffix, True


@dataclass
class TruncationResult:
    text: str
    orig_text: str
    head_lines_count: int
    tail_lines_count: int
    hidden_lines_count: int
    hidden_chars_count: int
    horiz_truncated_line_indices: List[int]

    @property
    def is_truncated(self) -> bool:
        return self.hidden_lines_count > 0 or bool(self.horiz_truncated_line_indices)

    def __iter__(self):
        # Support unpacking into (text, is_truncated)
        yield self.text
        yield self.is_truncated


LINE_SEP = "\n"


def split_text_lines(text: str) -> list[str]:
    # lines = text.split(LINE_SEP)
    lines = text.splitlines(keepends=True)
    return lines


def merge_text_lines(lines: list[str]) -> str:
    # text = LINE_SEP.join(lines)
    text = "".join(lines)
    return text


def truncate_text(
    orig_text: str, max_total_len: int, max_line_len: Optional[int] = None
) -> TruncationResult:
    # 1. Check total length first
    if len(orig_text) <= max_total_len:
        orig_lines = split_text_lines(orig_text)
        return TruncationResult(
            text=orig_text,
            orig_text=orig_text,
            head_lines_count=len(orig_lines),
            tail_lines_count=0,
            hidden_lines_count=0,
            hidden_chars_count=0,
            horiz_truncated_line_indices=[],
        )

    # 2. Horizontal Truncation (per line)
    orig_lines = split_text_lines(orig_text)
    horiz_processed_lines = []
    horiz_truncated_line_flags = []
    total_len_after_horiz = 0
    for line in orig_lines:
        if max_line_len is not None:
            line, truncated_flag = middle_truncate(line, max_line_len, False)
        else:
            truncated_flag = False
        horiz_processed_lines.append(line)
        horiz_truncated_line_flags.append(truncated_flag)
        total_len_after_horiz += len(line)
    horiz_processed_lines_count = len(horiz_processed_lines)

    # 3. Check if vertical truncation is needed after horizontal processing
    if total_len_after_horiz <= max_total_len:
        return TruncationResult(
            text=merge_text_lines(horiz_processed_lines),
            orig_text=orig_text,
            head_lines_count=horiz_processed_lines_count,
            tail_lines_count=0,
            hidden_lines_count=0,
            hidden_chars_count=0,
            horiz_truncated_line_indices=[
                i for i, t in enumerate(horiz_truncated_line_flags) if t
            ],
        )

    # 4. Calculate vertical cuts (Line-based to avoid cutting inside a line)
    head_threshold = max_total_len // 2
    tail_threshold = max_total_len - head_threshold

    # i_head: find the first line that ends at or exceeds head_threshold
    # We include this entire line in the head part.
    i_head = -1
    head_len = 0
    for i, line in enumerate(horiz_processed_lines):
        head_len += len(line)
        if head_len >= head_threshold:
            i_head = i
            break

    # i_tail: find the first line (from end) that makes the tail reach or exceed tail_threshold
    # We include this entire line in the tail part.
    i_tail = horiz_processed_lines_count
    tail_len = 0
    for i in range(horiz_processed_lines_count - 1, -1, -1):
        tail_len += len(horiz_processed_lines[i])
        if tail_len >= tail_threshold:
            i_tail = i
            break

    hidden_lines_count = i_tail - i_head - 1

    # 5. Check if we still need to truncate vertically
    # If head and tail sections overlap or touch, no lines are hidden.
    if hidden_lines_count <= 0:
        return TruncationResult(
            text=merge_text_lines(horiz_processed_lines),
            orig_text=orig_text,
            head_lines_count=horiz_processed_lines_count,
            tail_lines_count=0,
            hidden_lines_count=0,
            hidden_chars_count=0,
            horiz_truncated_line_indices=[
                i for i, t in enumerate(horiz_truncated_line_flags) if t
            ],
        )

    # 6. Construct Final Result (Line-based)
    head_part = horiz_processed_lines[: i_head + 1]
    tail_part = horiz_processed_lines[i_tail:]
    head_text = merge_text_lines(head_part)
    tail_text = merge_text_lines(tail_part)

    hidden_chars_count = total_len_after_horiz - head_len - tail_len
    marker_line = (
        f"...[TRUNCATED {hidden_lines_count} lines, {hidden_chars_count} chars]...\n"
    )
    final_text = head_text + marker_line + tail_text

    # 7. Map Truncated Indices
    truncated_line_indices = []

    # Head part indices (0 to i_head)
    for k in range(0, i_head + 1):
        if horiz_truncated_line_flags[k]:
            truncated_line_indices.append(k)

    # Tail part indices (i_tail to end)
    # We use negative indexing relative to the end of the final text.
    # -1 is the last line, -2 is the second to last, etc.
    for k in range(i_tail, horiz_processed_lines_count):
        if horiz_truncated_line_flags[k]:
            truncated_line_indices.append(k - horiz_processed_lines_count)

    return TruncationResult(
        text=final_text,
        orig_text=orig_text,
        head_lines_count=len(head_part),
        tail_lines_count=len(tail_part),
        hidden_lines_count=hidden_lines_count,
        hidden_chars_count=hidden_chars_count,
        horiz_truncated_line_indices=truncated_line_indices,
    )



from __future__ import annotations

import inspect
import json
import traceback
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable, Sequence
from enum import StrEnum
from functools import lru_cache
from typing import Self, cast

from judy_core.types import ContentUnion, Promptable

# from jsonschema.exceptions import ValidationError
from openai.types.chat import ChatCompletionFunctionToolParam
from openai.types.shared_params import FunctionDefinition
from pydantic import BaseModel, TypeAdapter
from pydantic import ValidationError as PydanticValidationError
from tongcell.util import getLogger

logger = getLogger(__name__)


@lru_cache(maxsize=5000)
def get_cached_typeadapter[T](cls: T) -> TypeAdapter[T]:
    """
    TypeAdapters are heavy objects, and in an application context we'd typically
    create them once in a global scope and reuse them as often as possible.
    However, this isn't feasible for user-generated functions. Instead, we use a
    cache to minimize the cost of creating them as much as possible.
    """
    return TypeAdapter(cls)


type JsonSchema = dict[str, object]


class ToolBasicDefinition(BaseModel):
    name: str
    description: str


class ToolDefinition(ToolBasicDefinition):
    parameters: JsonSchema

    def to_openai_format(self) -> ChatCompletionFunctionToolParam:
        return ChatCompletionFunctionToolParam(
            type="function",
            function=FunctionDefinition(
                name=self.name,
                description=self.description,
                parameters=self.parameters,
            ),
        )


def to_openai_format(
    tools_definition: Sequence[ToolDefinition],
) -> list[ChatCompletionFunctionToolParam]:
    return [tool_def.to_openai_format() for tool_def in tools_definition]


TOOL_RESULT_SUCCESS_CONTENT = "success"


class ToolErrorKind(StrEnum):
    NOT_FOUND = "not_found"
    INVALID_ARGS = "invalid_args"
    EXEC_FAIL = "tool_exec_fail"


class ToolErrorBase(ABC, Exception):
    kind: ToolErrorKind

    @abstractmethod
    def detail_message(self) -> str: ...


class ToolNotFound(ToolErrorBase):
    kind = ToolErrorKind.NOT_FOUND

    def __init__(self, name: str):
        super().__init__(f"tool not found:{name}")
        self.name = name

    def detail_message(self) -> str:
        return str(self)


class InvalidToolArgs(ToolErrorBase):
    kind = ToolErrorKind.INVALID_ARGS

    def __init__(self, ex: Exception):
        super().__init__(f"invalid tool args: {str(ex)}")
        self.ex = ex

    def detail_message(self) -> str:
        return str(self) + "\n" + "\n".join(traceback.format_exception(self.ex))


class ToolRuntimeError(ToolErrorBase):
    kind = ToolErrorKind.EXEC_FAIL

    def __init__(self, ex: Exception):
        super().__init__(f"tool runtime error: {str(ex)}")
        self.ex = ex

    def detail_message(self) -> str:
        return str(self) + "\n" + "\n".join(traceback.format_exception(self.ex))


class ToolExecResult:
    def __init__(self, content: ContentUnion, error: ToolErrorBase | None = None):
        self.content = content
        self.error = error

    @classmethod
    def success(cls, content: ContentUnion) -> Self:
        return cls(content=content, error=None)

    @classmethod
    def fail(cls, tool_error: ToolErrorBase) -> Self:
        text = "type: {}, msg:{}".format(type(tool_error).__name__, tool_error)
        return cls(
            content=text,
            error=tool_error,
        )

    def is_ok(self) -> bool:
        return self.error is None


def try_fix_model_args(args: str):
    decoded_args = json.loads(args)
    if isinstance(decoded_args, dict):
        for k, v in decoded_args.items():
            if isinstance(v, str):
                try:
                    decoded_args[k] = json.loads(v)
                except Exception:
                    pass
    return decoded_args


async def call_single_tool[T](tool: Callable[..., Awaitable[T]], args: str) -> T:
    type_adapter = get_cached_typeadapter(tool)
    try:
        # jsonschema.validate(dict_args, type_adapter.json_schema())
        coro = type_adapter.validate_json(args)
    except PydanticValidationError as e:
        logger.warning("tool args validate failed, trying autofix")
        try:
            fixed_args = try_fix_model_args(args)
            coro = type_adapter.validate_python(fixed_args)
        except Exception:
            raise InvalidToolArgs(e) from None

    coro = cast(Awaitable[T], coro)
    try:
        result_content = await coro
    except Exception as e:
        raise ToolRuntimeError(e)
    return result_content


def camel_to_snake(name: str) -> str:
    result = []

    n = len(name)
    for i, c in enumerate(name):
        if c.isupper():
            if i > 0:
                # 如果前一个字符或者后一个字符是小写，添加下划线
                if name[i - 1].islower() or (i + 1 < n and name[i + 1].islower()):
                    result.append("_")
            result.append(c.lower())
        else:
            result.append(c)

    return "".join(result)


type AsyncToolCallable[R] = Callable[..., Awaitable[R]]


class Tool[R]:
    def __init__(
        self, function: AsyncToolCallable[R], basic_definition: ToolBasicDefinition
    ):
        self._function = function
        self._definition = ToolDefinition(
            name=basic_definition.name,
            description=basic_definition.description,
            parameters=Tool.get_function_parameters(function),
        )

    def get_function(self) -> AsyncToolCallable[R]:
        return self._function

    def get_definition(self) -> ToolDefinition:
        return self._definition

    @staticmethod
    def get_function_parameters(function: AsyncToolCallable) -> JsonSchema:
        return get_cached_typeadapter(function).json_schema()

    @staticmethod
    def from_callable[R2](
        callable: AsyncToolCallable[R2],
        name: str | None = None,
        description: str | None = None,
    ) -> Tool[R2]:
        if inspect.iscoroutinefunction(callable):
            tool_function = callable
            default_name = callable.__name__
            default_description = callable.__doc__ or ""
        else:
            tool_function = callable.__call__
            callable_cls = type(callable)
            default_name = camel_to_snake(callable_cls.__name__)
            default_description = tool_function.__doc__ or ""

        tool_basic_def = ToolBasicDefinition(
            name=name or default_name,
            description=description or default_description,
        )

        return Tool(tool_function, tool_basic_def)


ToolReturnType = ContentUnion | Promptable


class ToolRegistry:
    def __init__(
        self,
        tools: Sequence[Tool[ToolReturnType] | AsyncToolCallable[ToolReturnType]],
    ):
        self._tool_dict = dict[str, Tool[ToolReturnType]]()
        for tool in tools:
            if not isinstance(tool, Tool):
                tool = Tool.from_callable(tool)
            self._tool_dict[tool.get_definition().name] = tool

    def get_tool_names(self) -> set[str]:
        return set(self._tool_dict.keys())

    def get_tools_definition(self) -> list[ToolDefinition]:
        definitions = []
        for name, tool in self._tool_dict.items():
            definitions.append(tool.get_definition())
        return definitions

    async def call_tool(self, name: str, args: str) -> ToolExecResult:
        tool = self._tool_dict.get(name)
        if tool is None:
            return ToolExecResult.fail(ToolNotFound(name))
        try:
            tool_result_content = await call_single_tool(tool.get_function(), args)
            if isinstance(tool_result_content, Promptable):
                tool_result_content = tool_result_content.to_prompt()
            return ToolExecResult.success(tool_result_content)
        except ToolErrorBase as tool_error:
            logger.warning(
                "⛔⛔⛔ name: %s, args: %s, error: %s",
                name,
                args,
                tool_error.detail_message(),
            )
            return ToolExecResult.fail(tool_error)



import traceback
from copy import copy
from typing import Any, Sequence, cast

import contexttimer
import litellm
from judy_core.agents.errors import AgentError, ModelBlankResponseError
from judy_core.agents.ids import CONFIGURATION_ACTOR_ID
from judy_core.events import (
    DebugModelInput,
    DebugModelOutput,
    DebugTaskLog,
    OutputEvent,
    TaskFailResult,
)
from judy_core.messages import GetModelConfig, GetTaskConfig, ModelConfig, TaskConfig
from judy_core.topics import publish_output_event
from judy_core.utils.llm import acompletion
from litellm.types.utils import Choices, Usage
from tongcell.core import (
    ActorStartMessage,
    MessageContext,
    RoutedActor,
    message_handler,
)
from tongcell.util import getLogger

logger = getLogger(__name__)


LLM_TOOLS_DISABLED_INSTRUCTION = """\
**Ignore this message. This message is only meant to remind you \
that you are now prohibited from calling any tools, and you should \
not give an empty response.**
"""


class LLMAgent(RoutedActor):
    @message_handler()
    async def on_start(self, msg: ActorStartMessage, ctx: MessageContext) -> None:
        self._model_config = await ctx.request(
            GetModelConfig(), CONFIGURATION_ACTOR_ID, response_type=ModelConfig
        )
        self._task_config = await ctx.request(
            GetTaskConfig(), CONFIGURATION_ACTOR_ID, response_type=TaskConfig
        )
        self._ctx = ctx

    async def on_handler_exception(
        self, message: Any, ex: Exception, ctx: MessageContext
    ) -> None:
        await self._publish_output_event(
            TaskFailResult(
                source=str(self.id), content="".join(traceback.format_exception(ex))
            )
        )
        return None

    def _need_explicit_cache(self) -> bool:
        need_explicit_cache = "claude" in self._model_config.model
        return need_explicit_cache

    async def _acompletion(
        self,
        messages,
        *,
        publish_model_output: bool = True,
        cache_control_injection_points: list | None = None,
        **kwargs,
    ) -> Choices:
        # logger.info(
        #     "messages %s", json.dumps(kwargs.get("messages"), ensure_ascii=False)
        # )
        if not self._need_explicit_cache():
            cache_control_injection_points = None
        with contexttimer.Timer() as t:
            litellm_completion_option = self._model_config.litellm_completion_option
            response = await acompletion(
                model=self._model_config.model,
                messages=messages,
                judy_option=self._model_config.judy_completion_option,
                judy_logger=self._debug_log,
                **litellm_completion_option.model_dump(),
                parallel_tool_calls=True,
                cache_control_injection_points=cache_control_injection_points,
                **kwargs,
            )
            if response is None:
                raise ModelBlankResponseError(
                    f"No choices returned from LLM: {response}"
                )

        choice = cast(Choices, response.choices[0])
        # TODO publish的消息可能还没来得及处理，runtime就出现异常把receiver stop了
        reasoning_content = getattr(choice.message, "reasoning_content", None)
        usage: Usage | None = getattr(response, "usage", None)

        cost_money = 0.0
        hidden_params = getattr(response, "_hidden_params", None)
        if isinstance(hidden_params, dict):
            response_cost = hidden_params.get("response_cost", 0.0)
            if response_cost:
                cost_money = response_cost

        if cost_money == 0.0:
            try:
                cost_money = litellm.completion_cost(completion_response=response)
            except Exception:
                pass

        if publish_model_output:
            await self._publish_output_event(
                DebugModelOutput(
                    source=str(self.id),
                    response_id=response.id,
                    model=response.model,
                    tool_calls=choice.message.tool_calls,
                    content=choice.message.content,
                    reasoning_content=reasoning_content,
                    usage=usage,
                    hidden_params=hidden_params,
                    cost_money=cost_money,
                    cost_time=t.elapsed,
                ),
            )

        return choice

    async def _acompletion_without_tool_calls(self, messages, **kwargs) -> Choices:
        messages = copy(messages)
        last_message = messages[-1]
        for i in range(0, 3):
            choice = await self._acompletion(messages, **kwargs)
            if choice.message.tool_calls or (not choice.message.content):
                await self._debug_log(
                    "agent {} try to call tools(or message.content is empty) while tool is disabled, choice:{}, last_message: {}".format(
                        self._id, choice, last_message
                    )
                )
                messages.append(
                    {
                        "role": "user",
                        "content": LLM_TOOLS_DISABLED_INSTRUCTION,
                    }
                )
                continue
            else:
                return choice
        error_content = "{} failed to generate a message without tool calls and text is not empty".format(
            self._id
        )
        await self._debug_log(error_content)
        raise AgentError(error_content)

    async def _publish_model_input(self, messages: Sequence) -> None:
        for message in messages:
            role = message.get("role")
            content = message.get("content")
            if isinstance(content, list):
                content = "".join([part.get("text", "") for part in content])
            if isinstance(role, str) and isinstance(content, str):
                await self._publish_output_event(
                    DebugModelInput(
                        source=str(self.id),
                        role=role,
                        content=content,
                    )
                )

    async def _publish_output_event(self, event: OutputEvent) -> None:
        await publish_output_event(self._ctx, event)

    async def _debug_log(self, text: str):
        await self._publish_output_event(
            DebugTaskLog(content="agent:{} {}".format(self._id, text))
        )



import asyncio
from collections.abc import AsyncGenerator

from judy_core.events import OutputEvent, TaskFailResult, TaskSuccessResult
from judy_core.topics import OUTPUT_TOPIC_KIND
from tongcell.core import (
    ActorStopMessage,
    RoutedActor,
    message_handler,
    subscribe_topic,
)


@subscribe_topic(OUTPUT_TOPIC_KIND)
class OutputStreamer(RoutedActor):
    def __init__(self, description: str = "", **kwargs) -> None:
        super().__init__(description, **kwargs)
        self._events = asyncio.Queue[OutputEvent]()

    @message_handler()
    async def on_stop(self, m: ActorStopMessage):
        # self._events.shutdown()
        pass

    @message_handler()
    async def _receive_output_event(self, event: OutputEvent):
        await self._events.put(event)

    async def stream(self) -> AsyncGenerator[OutputEvent]:
        try:
            while True:
                event = await self._events.get()
                yield event
                if isinstance(event, TaskSuccessResult | TaskFailResult):
                    return
        except asyncio.QueueShutDown:
            return



from judy_core.agents.llm_agent import LLMAgent
from judy_core.agents.types import ModelInputMessage


class StatefulLLMAgent(LLMAgent):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._messages: list[ModelInputMessage] = []
        self._tools = []



from judy_core.messages import (
    EnvironmentNotification,
    EnvironmentNotificationList,
    GetEnvironmentNotification,
)
from tongcell.core import MessageContext, RoutedActor, message_handler


class NotificationCenter(RoutedActor):
    def __init__(self):
        super().__init__()
        self._notifications = EnvironmentNotificationList()

    @message_handler()
    async def stage_notification(
        self, notification: EnvironmentNotification, ctx: MessageContext
    ) -> None:
        self._notifications.notifications.append(notification)

    @message_handler()
    async def get_notifications(
        self, msg: GetEnvironmentNotification
    ) -> EnvironmentNotificationList:
        notifications, self._notifications = (
            self._notifications,
            EnvironmentNotificationList(),
        )
        return notifications



from copy import copy

from judy_core.agents.stateful_llm_agent import StatefulLLMAgent
from judy_core.agents.types import ModelInputMessage
from judy_core.messages import MemberAnswer, MemberQuestion, TeamMemberName
from tongcell.core import MessageContext, message_handler

MEMBER_QA_INSTRUCTION_TEMPLATE = """\
从现在开始你进入与团队成员({question_member_name})的问答流程，对方提问，你来回答(每条回复不超过8000字，超过后中间部分会被截断，开头和尾部会被保留)。
注意：**在问答流程中，你不能调用任何工具**。
"""


class QAAgent(StatefulLLMAgent):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._qa_conversations: dict[TeamMemberName, list[ModelInputMessage]] = {}

    @message_handler()
    async def answer_question(
        self, question: MemberQuestion, ctx: MessageContext
    ) -> MemberAnswer:
        question_member_name = ctx.envelope.sender.name
        qa_messages = self._qa_conversations.setdefault(
            question_member_name,
            [
                {
                    "role": "user",
                    "content": MEMBER_QA_INSTRUCTION_TEMPLATE.format(
                        question_member_name=question_member_name
                    ),
                }
            ],
        )
        qa_messages.append({"role": "user", "content": question.text})

        messages = copy(self._messages)
        messages.extend(qa_messages)
        choice = await self._acompletion_without_tool_calls(
            messages=messages,
            tools=self._tools,
            publish_model_output=False,
        )
        qa_messages.append(choice.message)

        assert choice.message.content is not None
        return MemberAnswer(text=choice.message.content)



from tongcell.core import ActorId, LocalRuntime

from judy_core.agents.configuration import Configuration
from judy_core.agents.executor import Executor
from judy_core.agents.notification_center import NotificationCenter
from judy_core.agents.output_streamer import OutputStreamer
from judy_core.agents.planner import Planner
from judy_core.agents.state_manager import TaskStateManager
from judy_core.agents.verifier import Verifier
from judy_core.messages import (
    ModelConfig,
    Task,
    TaskConfig,
)


async def create_task_runtime(
    task: Task,
    task_config: TaskConfig,
    model_config: ModelConfig,
) -> tuple[LocalRuntime, OutputStreamer]:
    runtime = LocalRuntime()
    for agent_cls in [Planner, Executor, Verifier, NotificationCenter]:
        await agent_cls.register(runtime)
    config = Configuration(task_config, model_config)
    await runtime.register_actor_factory(Configuration.actor_kind, lambda: config)
    state_manager = TaskStateManager()
    await runtime.register_actor_factory(
        TaskStateManager.actor_kind, lambda: state_manager
    )
    output_streamer = OutputStreamer()
    await runtime.register_actor_factory(
        OutputStreamer.actor_kind, lambda: output_streamer
    )
    await OutputStreamer.register_topic_subscription(runtime)

    await runtime.send(task, ActorId.default(Planner.actor_kind))
    return runtime, output_streamer



import os

index = 0


def get_model_name():
    model_name_env_value = os.getenv("JUDY_MODEL", "")
    if not model_name_env_value:
        raise RuntimeError(
            "Please set env var `JUDY_MODEL` (or create a `.env` file) "
            "to a valid litellm model name, e.g. `openai/gpt-4.1` or `vertex_ai/gemini-3-pro`."
        )
    model_names = model_name_env_value.split(",")

    if len(model_names) > 1:
        # import random
        # model_name = random.choice(model_names)
        global index
        if index >= len(model_names):
            index = index % len(model_names)
        model_name = model_names[index]
        index = index + 1
    else:
        model_name = model_name_env_value

    return model_name



import time
from copy import copy
from typing import Annotated

from judy_core.agents.ids import VERIFIER_NAME, get_executor_name
from judy_core.agents.llm_agent import LLMAgent
from judy_core.agents.prompts import AGENT_COMMON_INSTRUCTION, AGENT_FOLDER_INSTRUCTION
from judy_core.agents.types import ModelInputMessage
from judy_core.agents.verifier import Verifier
from judy_core.events import DebugTaskLog, Task, TaskFailResult, TaskSuccessResult
from judy_core.messages import (
    ExecutionEndReason,
    ExecutorResult,
    Plan,
    Subtask,
    TaskHistory,
    Todo,
    TodoStatus,
    VerificationRequest,
    VerificationResult,
    VerificationResultStatus,
)
from judy_core.utils.messages import convert_content_to_openai_format
from judy_core.utils.tool import Tool, ToolRegistry
from litellm import BaseModel, Choices
from litellm.types.completion import ChatCompletionToolMessageParam
from pydantic import Field
from tongcell.core import ActorId, MessageContext, message_handler
from tongcell.util import getLogger

SAVE_PLAN_TOOL_NAME = "save_plan"
PLANNER_INSTRUCTION = f"""你是一个编程任务的planner，你与团队中的一组executor协作来尝试完成一项任务。

# 你的角色和职责
你是团队中的架构师（CTO），负责制定规划（plan）。规划由一系列todo构成，你需要保证每个todo清晰、可行，并调用`{SAVE_PLAN_TOOL_NAME}`工具来保存规划。
系统会自动为第一个处于PENDING状态的todo分配一个executor，executor执行该todo，然后将结果反馈给你。你需要根据反馈更新todo的状态、调整规划。
以上过程重复进行，直到你认为任务已经完成，然后你将所有todo标记为完成状态，并做一个总结。

## 不必过度拘泥于执行细节
你主要专注于把控（调整）规划的整体方向和进度，确保规划完整、可行，能够解决分配给团队的任务。
在设计/调整规划时，不必过度纠结于todo的执行细节，而是应该把这些细节交给executor去解决。因为你没有与任务环境交互的能力，只能空想，而executor可以边思考边尝试，在这方面要比你效率更高。

# executor

## executor的能力
executor可以执行的操作包括：
- 向其他executor, verifier提问
- 运行命令
- 与shell终端交互
- 读写文件
- 使用自身的视觉能力，读取、理解、处理部分多媒体文件（图片、pdf）
- 网页搜索和抓取
- 调用LLM api（支持多模态）进行批处理。llm调用成本较高，因此最好仅对抽样出的（或者可疑的）数据/结果进行分析）

## executor的记忆相互隔离
executor并不知道其他executor做了什么(除非其主动向其他executor提问)，所以你给executor的todo中最好包含执行该todo所依赖的信息。
如果缺乏相关信息，executor会自行调用工具去重新探索或者询问其它executor，但这样会导致效率降低，任务耗时增加，有任务超时的风险。

## executor可以向你提建议
executor对他自己的能力更为了解，同时他们工作在一线，获取到的信息更加具体。因此executor可能根据他的发现，向你建议更优的方案。
而你(planner)相比executor有更全局的任务视角优势，因此executor的建议也不一定正确，你需要综合任务历史和现状做出最优决策。

# 规划

## todo的粒度
- todo不宜过于简单，每个todo预期的executor执行步数应该至少为3,否则会增加与executor的协调成本，降低效率；
- todo也不宜过于复杂，否则executor可能因为过于复杂或者困难而无法在规定的步数内完成。

## 任务描述中可能会缺乏环境信息
在缺乏环境信息时，制定的规划可能会严重偏离任务原本的方向/目的，导致任务耗时增加，甚至任务失败。
因此当任务没有明确描述环境信息时（如当前目录下的文件、操作系统版本等），你需要先生成一个探索环境的规划，然后根据探索得到的信息，再生成解决任务的规划。

## 遵循最佳实践
收到任务后，要先分析任务所属的类别/领域，然后根据该领域的最佳实践来进行规划。
如：
任务：从xxx中恢复出yyy数据。
分析：该任务属于`数据恢复`领域，而数据恢复的最佳实践/准则是：在进行任何恢复操作之前，停止任何写入、立即进行备份。遵循该最佳实践可以防止操作意外导致数据损毁/丢失。
因此规划中的第1个todo必须是`数据备份`，其后才是`探索环境`等todo。

## 每项TODO完成后，要将其状态改为COMPLETED，而不是直接从plan中删除
由于每项todo是由不同的executor执行，删除todo会导致executor无法看到完整的规划，对其负责的todo感到困惑， 从而可能重复执行被，或者不遵守你的指令，自由发挥。
一旦出现这种局面，极易导致任务失败。
因此你在调整规划时，切记**保持规划的完整性，如果某项TODO是当前求解规划所必须的，且已经完成了，不要删除它**。
但如果你将规划推倒重来，或者大幅度调整了规划方向，导致某些TODO在新的规划中不再适用，那么你仍然可以自由地删除这些TODO。

{AGENT_FOLDER_INSTRUCTION}

{AGENT_COMMON_INSTRUCTION}
"""

logger = getLogger(__name__)


class TurnResult(BaseModel):
    ok: bool = True
    msg: str = ""

    @classmethod
    def success(cls, msg="") -> "TurnResult":
        return cls(ok=True, msg=msg)

    @classmethod
    def fail(cls, msg) -> "TurnResult":
        return cls(ok=False, msg=msg)


class PromptCacheInfo(BaseModel):
    index: int
    timestamp: float

    @classmethod
    def create(cls, index: int) -> "PromptCacheInfo":
        return cls(index=index, timestamp=time.perf_counter())

    def is_expired(self, duration: float) -> bool:
        now = time.perf_counter()
        return now - self.timestamp >= duration


class Planner(LLMAgent):
    def __init__(self):
        super().__init__()
        self._tool_registry = ToolRegistry(
            [Tool.from_callable(self._save_plan_tool, SAVE_PLAN_TOOL_NAME)]
        )
        self._tools = [
            tool_def.to_openai_format()
            for tool_def in self._tool_registry.get_tools_definition()
        ]

        self._messages: list[ModelInputMessage] = []
        self._prompt_cache_info: PromptCacheInfo | None = None
        self._task_history: TaskHistory = TaskHistory(team_member_results=[])
        self._failed_plans = list[Plan]()
        self._plan: Plan | None = None
        self._task_execution_summary = ""

        self._executor_id_counter = 0

    def next_executor_name(self) -> str:
        executor_name = get_executor_name(str(self._executor_id_counter))
        self._executor_id_counter += 1
        return executor_name

    async def _save_plan_tool(
        self,
        task_finished: Annotated[
            bool, Field(description="任务完成标记。在无需进一步规划时设置为为true。")
        ],
        todos: list[Todo],
    ) -> str:
        """必须提供完整的规划，完整是指没有缺少必要的步骤(todo)"""
        plan = Plan(task_finished=task_finished, todos=todos)
        self._plan = plan

        if not self._plan.task_finished:
            return "success."

        pending_todo_count = sum(
            todo.status == TodoStatus.PENDING for todo in self._plan.todos
        )
        if pending_todo_count != 0:
            raise RuntimeError(
                "错误：状态不一致，task_finished为true时，规划中不能包含处于PENDING状态的todo。"
            )
        return "success. " + "你已经将任务标记为完成，现在请生成任务总结。"

    @message_handler()
    async def start_task(self, task: Task, ctx: MessageContext) -> None:
        self._task = task
        planner_instruction = PLANNER_INSTRUCTION
        if task.env_notes:
            planner_instruction += task.env_notes
        prompt = "\n".join(
            [
                self._task.to_prompt(),
                self._task_config.to_prompt(),
            ]
        )
        self._messages = [
            {"role": "system", "content": planner_instruction},
            {"role": "user", "content": prompt},
        ]
        await self._publish_model_input(self._messages)
        await self._do_turn(ctx)

    @message_handler()
    async def receive_executor_result(
        self, executor_result: ExecutorResult, ctx: MessageContext
    ) -> None:
        self._task_history.team_member_results.append(executor_result)

        assert self._plan is not None
        finished_todo = self._plan.todos[executor_result.todo_index]
        content = (
            f"executor(名字：{executor_result.member_id.name})执行期间花费了{executor_result.step_count}步，调用了{executor_result.tool_call_count}次工具。\n"
            f"其执行了当前plan中的第{executor_result.todo_index + 1}项todo："
            f"<todo>{finished_todo.desc}。<todo>\n"
            f"其执行的最终状态是：{executor_result.end_reason.str_for_model()}\n"
        )
        if executor_result.extra_finished_todo_indexes:
            indexes_for_model = [
                index + 1 for index in executor_result.extra_finished_todo_indexes
            ]
            content += f"其还认为额外完成了第{indexes_for_model}项todo。\n"
        if executor_result.end_reason == ExecutionEndReason.STEPS_EXHAUSTED:
            content += (
                "注意：其因为超过了最大步数被中止了执行，被要求就取得的进展进行总结。\n"
            )
        content += f"executor的汇报（反馈）如下：<report>{executor_result.report_content}</report>\n"
        message: ModelInputMessage = {
            "role": "user",
            "content": content,
        }
        self._messages.append(message)
        await self._do_turn(ctx)

    async def _do_turn(self, ctx: MessageContext) -> None:
        LIMIT = 6
        for i in range(0, LIMIT):
            turn_result = await self._do_turn_inner(ctx)
            if turn_result.ok:
                return
            else:
                await self._debug_log(
                    "replanning {}/{}, reason: {}".format(i + 1, LIMIT, turn_result.msg)
                )
        await self._publish_output_event(
            TaskFailResult(content="planner stuck in do_turn loop")
        )

    async def _do_turn_inner(self, ctx: MessageContext) -> TurnResult:
        messages = copy(self._messages)
        if self._need_explicit_cache():
            last_message = copy(messages[-1])
            messages[-1] = last_message
            if isinstance(last_message, dict):
                ttl = None
                if self._plan:
                    team_member_result = self._task_history.team_member_results[-1]
                    if (
                        isinstance(team_member_result, ExecutorResult)
                        and team_member_result.end_reason
                        == ExecutionEndReason.COMPLETED
                    ):
                        pending_todo_count = sum(
                            todo.status == TodoStatus.PENDING
                            for todo in self._plan.todos
                        )
                        if len(self._plan.todos) > 1 and pending_todo_count == 1:
                            ttl = "5m"
                if ttl:
                    cache_control = {
                        "type": "ephemeral",
                        "ttl": ttl,
                    }
                    last_message["cache_control"] = cache_control  # pyright: ignore[reportGeneralTypeIssues]
                    self._prompt_cache_info = PromptCacheInfo.create(len(messages) - 1)

        choice = await self._acompletion(
            messages=messages,
            tools=self._tools,
        )
        tool_calls = choice.message.tool_calls

        if not tool_calls:
            self._messages.append(
                {
                    "role": "user",
                    "content": f"Ignore this message; it is only intended to prompt you to call the {SAVE_PLAN_TOOL_NAME} tool.",
                }
            )
            return TurnResult.fail("plan not generated")

        if len(tool_calls) > 1:
            return TurnResult.fail("multiple plan generated")

        tool_call = tool_calls[0]
        if tool_call.function.name is None:
            return TurnResult.fail("function name is empty")

        tool_exec_result = await self._tool_registry.call_tool(
            tool_call.function.name,
            tool_call.function.arguments,
        )
        self._messages.append(choice.message)
        self._messages.append(
            ChatCompletionToolMessageParam(
                content=convert_content_to_openai_format(tool_exec_result.content),
                tool_call_id=tool_call.id,
                role="tool",
            )
        )
        if not tool_exec_result.is_ok():
            return TurnResult.fail("save_plan tool call failed")

        if not self._plan or len(self._plan.todos) == 0:
            await self._publish_output_event(
                DebugTaskLog(content="planner: empty plan")
            )
            self._messages.append(
                {
                    "role": "user",
                    "content": "规划不能为空",
                }
            )
            return TurnResult.fail("plan is empty")

        for i, todo in enumerate(self._plan.todos):
            if todo.status == TodoStatus.PENDING:
                from judy_core.agents.executor import Executor

                await ctx.send(
                    Subtask(
                        task=self._task,
                        task_history=self._task_history.model_copy(deep=True),
                        plan=self._plan,
                        todo_index=i,
                    ),
                    ActorId(Executor.actor_kind, self.next_executor_name()),
                )
                return TurnResult.success("handle over to executor")
        if not self._plan.task_finished:
            return TurnResult.fail(
                "all todo of plan are done, but task_finished flag is not set"
            )

        choice = await self._gen_task_summary()
        assert choice.message.content is not None
        self._task_execution_summary = choice.message.content

        if self._task_config.disable_verifier:
            await self._publish_output_event(
                TaskSuccessResult(content=self._task_execution_summary)
            )
        else:
            await ctx.send(
                VerificationRequest(
                    task=self._task, task_execution_summary=self._task_execution_summary
                ),
                ActorId(Verifier.actor_kind, VERIFIER_NAME),
            )
        return TurnResult.success("task done")

    @message_handler()
    async def receive_verification_response(
        self, verification_result: VerificationResult, ctx: MessageContext
    ):
        self._task_history.team_member_results.append(verification_result)

        if verification_result.status == VerificationResultStatus.FAILED:
            # TODO 要求重新规划
            self._messages.append(
                {
                    "role": "user",
                    "content": f"任务验收没有通过，验收报告: <verify_result status={verification_result.status}>{verification_result.report_content}</verify_result>。",
                }
            )
            await self._do_turn(ctx)
        else:
            self._messages.append(
                {
                    "role": "user",
                    "content": f"任务验收通过，验收报告: <verify_result>{verification_result.report_content}</verify_result>。",
                }
            )
            await self._publish_output_event(
                TaskSuccessResult(content=self._task_execution_summary)
            )
        return

    async def _gen_task_summary(self) -> Choices:
        cache_control_injection_points = None
        if self._prompt_cache_info and not self._prompt_cache_info.is_expired(5 * 60):
            cache_control_injection_points = [
                {
                    "location": "message",
                    "index": self._prompt_cache_info.index,
                }
            ]
        choice = await self._acompletion_without_tool_calls(
            messages=self._messages,
            tools=self._tools,
            cache_control_injection_points=cache_control_injection_points,
        )
        self._messages.append(choice.message)
        return choice



from judy_core.messages import GetTaskState, TaskState
from tongcell.core import MessageContext, RoutedActor, message_handler


class TaskStateManager(RoutedActor):
    def __init__(self):
        super().__init__()
        self._task_state = TaskState()

    @message_handler()
    async def get_task_state(self, _: GetTaskState, ctx: MessageContext) -> TaskState:
        return self._task_state



from abc import ABC, abstractmethod

import contexttimer
from judy_core.agents.errors import AgentError
from judy_core.agents.ids import NOTIFICATION_CENTER_ACTOR_ID
from judy_core.agents.stateful_llm_agent import StatefulLLMAgent
from judy_core.events import DebugToolResult, TaskFailResult, ToolCall, ToolResult
from judy_core.messages import EnvironmentNotificationList, GetEnvironmentNotification
from judy_core.utils.messages import convert_content_to_openai_format
from judy_core.utils.tool import ToolRegistry
from litellm.types.completion import (
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from tongcell.core import MessageContext, message_handler


class ReactAgent(StatefulLLMAgent, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._internal_tool_registry = ToolRegistry(tools=[])

        self._step_count = 0
        self._tool_call_count = 0

        self._tool_calls = []
        self._tool_results = dict[str, ToolResult]()

    @abstractmethod
    def _react_exit_prompt(self) -> str: ...

    @abstractmethod
    async def _on_turn_done(self, ctx: MessageContext) -> None: ...

    async def _do_turn(self, ctx: MessageContext) -> None:
        for i in range(0, 5):
            ok = await self._do_turn_inner(ctx)
            if ok:
                return
        await self._publish_output_event(
            TaskFailResult(content="executor stuck in do_turn loop")
        )

    async def _do_turn_inner(self, ctx: MessageContext) -> bool:
        await self._add_notification_message(ctx)

        choice = await self._acompletion(
            messages=self._messages,
            tools=self._tools,
            cache_control_injection_points=[
                {
                    "location": "message",
                    "index": -1,
                }
            ],
        )
        if choice.message.tool_calls:
            self._tool_message = choice.message
            self._tool_calls = choice.message.tool_calls
            self._tool_results.clear()
            internal_tool_names = self._internal_tool_registry.get_tool_names()
            for tool_call in choice.message.tool_calls:
                tool_call_id = tool_call.id
                function_name = tool_call.function.name
                function_arguments = tool_call.function.arguments
                if not function_name:
                    raise AgentError("empty function name")
                if function_name in internal_tool_names:
                    with contexttimer.Timer() as t:
                        tool_exec_result = await self._internal_tool_registry.call_tool(
                            function_name, function_arguments
                        )
                    await self._publish_output_event(
                        DebugToolResult(
                            source=str(self._id),
                            tool_call_id=tool_call_id,
                            name=function_name,
                            content=tool_exec_result.content,
                            time_cost=t.elapsed,
                        )
                    )
                    await ctx.send(
                        ToolResult(
                            id=tool_call_id,
                            name=function_name,
                            content=tool_exec_result.content,
                        ),
                        self._id,
                    )
                else:
                    await self._publish_output_event(
                        ToolCall(
                            source=str(self._id),
                            tool_call_id=tool_call_id,
                            name=function_name,
                            arguments=function_arguments,
                        ),
                    )
            return True
        else:
            await self._debug_log(text="did not issue any tool calls during react step")
            self._messages.append(
                {
                    "role": "user",
                    "content": self._react_exit_prompt(),
                }
            )
            return False

    async def _add_notification_message(self, ctx: MessageContext):
        notification_list = await ctx.request(
            GetEnvironmentNotification(),
            NOTIFICATION_CENTER_ACTOR_ID,
            response_type=EnvironmentNotificationList,
        )
        if not notification_list.notifications:
            return
        text_segs = ["There are some notifications for you. "]
        for notification in notification_list.notifications:
            text_segs.extend(
                [
                    "<notification>",
                    notification.text,
                    "</notification>",
                ]
            )
        message = ChatCompletionUserMessageParam(
            role="user", content="\n".join(text_segs)
        )
        await self._publish_model_input([message])
        self._messages.append(message)

    @message_handler()
    async def receive_tool_result(
        self, tool_result: ToolResult, ctx: MessageContext
    ) -> None:
        id_found = False
        for tool_call in self._tool_calls:
            if tool_call.id == tool_result.id:
                id_found = True
                break
        if not id_found:
            raise AgentError(f"tool_call id not found: {tool_result.id}")
        if self._tool_results.get(tool_result.id) is not None:
            raise AgentError(f"tool_call result already set: {tool_result.id}")
        self._tool_results[tool_result.id] = tool_result
        self._tool_call_count += 1

        if len(self._tool_calls) > len(self._tool_results):
            return None

        tool_turn_messages: list = [self._tool_message]
        media_content_parts = []
        for tool_call in self._tool_calls:
            temp_tool_result = self._tool_results[tool_call.id]
            tool_result_content = convert_content_to_openai_format(
                temp_tool_result.content, self._model_config.model
            )

            if (
                "claude" in self._model_config.model
                and "openrouter" in self._model_config.model
            ) or "gemini" in self._model_config.model:
                # openroute上的claude api会过调掉工具返回的多模态内容。
                # gemini 3.0虽然支持工具返回多模态内容，但litellm在转换时将工具返回的内容转换为了多个part，没有转换成gemini 3.0需要的functionResponse part的子part。
                # 导致gemini经常出现乱码，message中出现reasoning内容，不断重复相同文本内容/tool call，tool call内容非法等诡异问题。
                if not isinstance(tool_result_content, str):
                    tool_result_text_content = []
                    for part in tool_result_content:
                        if part.get("type") == "text":
                            tool_result_text_content.append(part)
                        else:
                            media_content_parts.append(
                                {
                                    "type": "text",
                                    "text": "The following content is the media file read by tool call `{} {}`".format(
                                        tool_call.function.name,
                                        tool_call.function.arguments,
                                    ),
                                }
                            )
                            media_content_parts.append(part)
                    tool_result_content = tool_result_text_content

            tool_turn_messages.append(
                ChatCompletionToolMessageParam(
                    tool_call_id=tool_call.id,
                    content=tool_result_content,
                    role="tool",
                )
            )

        self._messages.extend(tool_turn_messages)
        if media_content_parts:
            self._messages.append(
                ChatCompletionUserMessageParam(
                    role="user",
                    content=media_content_parts,
                )
            )

        await self._on_turn_done(ctx)



from pydantic import BaseModel


class JudyCompletionOption(BaseModel):
    stream_chunk_timeout: float | None = 160
    num_retries: int = 5


class LitellmCompletionOption(BaseModel):
    # litellm默认超时时间是10分钟，太长了
    timeout: float = 150
    num_retries: int = 5



from tongcell.core import MessageContext, TopicId, TopicKind

from judy_core.events import OutputEvent

OUTPUT_TOPIC_KIND = TopicKind("output")
OUTPUT_TOPIC_ID = TopicId.default(OUTPUT_TOPIC_KIND)
TRAJECTORY_TOPIC_KIND = TopicKind("trajectory")


async def publish_output_event(ctx: MessageContext, event: OutputEvent):
    await ctx.publish(event, OUTPUT_TOPIC_ID)



from judy_core.paths import JUDY_DIR, SPACE_DIR_PREFIX, SPACE_SHARED_DIR

AGENT_COMMON_INSTRUCTION = """\
# 全体要求（适用于所有人）

## 代码规范
除非任务明确要求，否则：
- 禁止使用废弃或者不再推荐使用的语法、api、sdk、代码范式。只要当前编译器/解释器支持，就应该使用最新、最推荐的、更容易理解、更安全的（不容易造成bug的）语法、api、sdk、代码范式。

"""

AGENT_FOLDER_INSTRUCTION = f"""\
# 目录和文件

## 目录介绍

### 任务目录
任务目录中包含任务所提供的文件。

### tool工作目录
tool的工作目录(cwd）默认情况下是任务目录。
如果tool支持，可以通过cwd参数来修改tool的工作目录。

### 交付目录
交付目录是用来保存任务要求交付的文件的目录。
一般情况下交付目录与任务目录相同，但如果任务中明确指定了交付文件的目录，则以任务说明为准。

### 团队空间（space）
团队空间的路径一般是`{JUDY_DIR}/{SPACE_DIR_PREFIX}`。
团队空间用来保存任务解决过程中所创建的文件。
每个团队成员都可以自由读写团队空间目录。

团队空间的目录结构
`{JUDY_DIR}/{SPACE_DIR_PREFIX}`
.
├── shared     团队共享目录
├── executor-0 团队成员executor-0的个人目录
├── executor-1 团队成员executor-1的个人目录
├── executor-xxx 团队成员executor-xxx的个人目录
└── verifier-0   团队成员verifier-0的个人目录


## 目录使用说明

### 交付目录必须保持整洁
只有任务明确要求交付/生成的文件，才能保存到交付目录。

**如果你不小心在交付目录创建了无关文件，在结束执行前必须转移到团队空间（推荐）或者删除。如果你没来得及清理就进入了汇报阶段，则必须在汇报中说明这个问题，以便其它人帮忙操作。**

任务解决过程中直接或者间接创建的的文件，如：
- 编写的用于debug、测试的代码、脚本
- 脚本/命令创建的文件
- 下载的文件
- 编译的产物(如：g++ -o)
- 备份文件
这些非交付文件禁止保存到交付目录，必须保存到团队空间或个人目录。

### 团队空间
如果文件需要被团队成员共享，则保存到团队空间的共享目录({SPACE_SHARED_DIR})。
否则应该保存到个人目录。

## 文件共享
文件共享对团队协作至关重要。 如：
- 保持汇报整洁、重点突出：将大块数据和代码拆分到文件中，报告中仅包含路径，避免污染报告内容。
- 后位(执行顺序靠后)的executor需要在前位的executor取得的进展（脚本、数据）的基础上继续工作。
- 后位的executor/verifier需要历史文件和数据，来复盘和定位问题根源。
- executor需要verifier的测试脚本来复现问题。
通过文件共享，能够实现信息共享和团队效率提升（避免重复编写代码/脚本）

## 用mv代替rm
- mv更安全，可以避免误删文件。
- 当任务进展不及预期时，执行过程中产生的历史文件（脚本、数据）是复盘的关键。
- 不论任务成功或者失败，历史文件对用户（任务下发者）都很有用。

因此**不要轻易删除历史文件，而是根据用途保存到合适的地方（团队空间、共享目录、个人目录）。**
"""



from enum import Enum
from itertools import chain
from typing import Annotated

from judy_core.agents.prompts import AGENT_COMMON_INSTRUCTION, AGENT_FOLDER_INSTRUCTION
from judy_core.agents.qa_agent import QAAgent
from judy_core.agents.react_agent import ReactAgent
from judy_core.agents.tools import ASK_MEMBER_TOOL_NAME, AskMemberTool, AskTargets
from judy_core.events import CreateAgentFolder
from judy_core.messages import (
    ExecutionEndReason,
    ExecutorResult,
    Subtask,
    TaskHistory,
    TeamMemberId,
)
from judy_core.utils.tool import TOOL_RESULT_SUCCESS_CONTENT, Tool, ToolRegistry
from pydantic import Field
from tongcell.core import ActorId, MessageContext, message_handler
from tongcell.util import getLogger

END_EXECUTION_TOOL_NAME = "end_execution"

EXECUTOR_NOT_ISSUE_ANY_TOOL_CALL_INSTRUCTION = f"""如果你认为当前已经可以进入汇报阶段，你需要调用{END_EXECUTION_TOOL_NAME}工具；否则你应该调用其他工具继续执行。"""

EXECUTOR_FEEDBACK_INSTRUCTION = "你需要就目前的发现和取得的进展进行汇报，汇报中可以包含你的建议。该汇报会被提交给planner，以便其及时调整规划。"

EXECUTOR_FEEDBACK_PHASE_INSTRUCTION = f"""\
你现在进入了汇报阶段，{EXECUTOR_FEEDBACK_INSTRUCTION}**你不能再调用任何工具了，直接将汇报作为你的回复内容即可**。\
"""

EXECUTOR_MAX_STEPS = 80
EXECUTOR_STEPS_EXAHUSTED_INSTRUCTION = f"你已经执行到了第{EXECUTOR_MAX_STEPS}步，超过了最大步数限制，你不能再继续了。{EXECUTOR_FEEDBACK_PHASE_INSTRUCTION}"


TASK_HISTORY_INSTRUCTION = f"""\
# 任务历史
任务历史由团队成员的执行结果组成，按照成员的执行顺序排列。

executor的执行结果主要包含其负责的todo，汇报内容等:
- step_count表示其执行步数，tool_call_count表示其工具调用次数，因为一步可以调用多个工具，所以tool_call_acount可能大于step_count。
- 部分executor会一次性完成多个todo。
- 由于planner会根据反馈调整规划，所以executor执行过的一些todo并不包含在当前规划中。

verifier的执行结果主要包含其验证报告。

## 善用提问
`{ASK_MEMBER_TOOL_NAME}` tool可用于向其他团队成员（任务历史中包含有团队成员的名字）提问。

提问的用途包括但不限于：
- 解决任务历史中团队成员的报告内容不够清晰，过于简略，或者由于内容过长被截断的问题
- 收集团队成员的成功/失败经验，避免重复工作，提高效率
- 在verifier给出验收失败的结论后，向其咨询验收的方式

注意：
- 执行步数/工具调用次数(step count > 10)越多的成员，向其提问价值越高，因为其汇报内容可能并没有完全覆盖他发现的所有信息和尝试过的所有路径。
- 团队成员都非常热心，会诚实和毫无保留地回答你的问题。
- 你可以通过一次性向多个人提问来提高效率。
"""


EXECUTOR_INSTRUCTION = f"""\
你的团队通过协作来尝试完成一项编程任务，团队由1个planner，1个verifier，和一组executor构成。你是其中一个executor。

# 对你的要求

## 分工
planner负责制定规划（由todo列表构成），为当前规划中第一个处于PENDING状态的todo分配executor，executor使用为其提供的工具来解决这个todo。
executor配备有相同的tool，像接力赛一样接力处理规划中的todo。你是这些executor其中之一。

## **你可以向planner提建议**
- 由于planner不知道你有哪些工具，以及你的工具的适用范围。可能也不知道作为一线人员的executor所发现的细节，所以planner给出的plan或者todo不一定是最优的。
- 如果根据你的系统指令、工具的说明、以及你所发现的信息，你认为当前规划/todo并不合理，或者你发现有更优/更合适的方案时，你可以中止执行，直接进行任务汇报，告诉planner你的想法和建议。
- 你的建议可以不局限于你当前所负责的todo，也可以包含已经执行过的todo或者整体规划的建议。

{AGENT_FOLDER_INSTRUCTION}

# 工具

## 并行工具调用
你可以在一次回复中包含多个工具调用（tool calls），这些工具调用会被依次执行。你必须充分利用此特性来减少交互轮数，降低任务耗时。
此外如非必要，不要使用shell的`&&`或`;`或`\n`语法来在一个工具调用中串联多个命令，而是应该拆分为多个工具调用，在一次回复中一次性返回。这样多个命令的输出不会混在一起，对你理解工具的执行结果更有帮助。

## 尽量通过并行来减少步数，减少任务耗时
除了并行工具调用外，你还可以通过其它方式（比如批量运行）来提高效率。比如你在编写脚本时，不要一个一个地去验证或者测试，而是编写一个包含多条数据或者用例的脚本，批量验证。
**在进行parallel tool call时，你必须确保生成调用llm的input文件的tool call，位于调用llm 的tool call之前。**

## 工具调用的超时时间
在运行命令或者等待命令退出时，超时时间最好设置在120秒以下，定期观察进度，有助于及时发现和处理死锁、卡顿等异常问题。
你可以在等待期间尝试其它方案，思考进程的耗时是否符合预期。

**如果代码运行很久都没有出结果，有可能是因为代码存在bug**，如：
- 死循环、无限递归、时间复杂度过高等计算量过大的问题
- 死锁类问题
你可以通过查看/proc目录下进程的cpu使用时间的变化来确定这类问题。

## 工具的输出可能会被截断
如果工具的输出内容过长，为了避免读取到大量垃圾内容，会被截断：
- 行内截断：当某行内容过长时，行中间部分会被截断，以 `...[TRUNCATED yyy chars]...` 来表示。
- 中间行截断：经过行内截断处理后，如果整体内容仍然过长，则内容头部、尾部各保留一部分，中间部分的行会被整体截断，以 `...[TRUNCATED xxx lines, yyy chars]...` 表示

### 截断处理策略
如果被截断的内容可能含有关键信息，那么你需要进行过滤，在减少无意义内容的同时保留关键信息。 比如：
- 对于行内截断：使用`grep -o`来仅输出匹配内容
- 对于中间行截断：使用`grep`,`grep-v`来保留和过滤整行内容。
如果简单的过滤命令无法满足需求，你可以编写脚本来实现复杂的过滤逻辑 。

**过滤容易导致意外错过重要信息**
- 过滤策略应该从宽松逐渐滑向严格，不要直接转向最严格的策略
- 不要只使用一种搜索/过滤策略，最好正向、反向、宽松、严格等多种策略都尝试一下
为了避免遗漏信息，你应该综合交叉验证后再下结论。

## 调试(debug)
排查bug（或者coredump）时，相比盲猜和不断地改代码加日志，通过debugger（如gdb、pdb）定位问题更加高效和准确。
可以通过在interactive shell中启动debugger来交互式地调试代码。

## 网页搜索
如果你拥有搜索网页的tool，或者拥有可以开启网页搜索功能调用外部llm的tool，你应该优先使用这些tool来搜索网页。
这些tool的搜索质量和效率一般情况下比你自行写代码爬取搜索页面要好很多，且没有限流、反爬虫等问题。

### 把搜索tool当成你的百科全书
**如果你对某些领域/知识/信息不是很了解，或者拿不准，不要瞎猜，务必通过搜索来补充相关知识。**

## 多模态
对于音视频、图片或者布局复杂/非结构化的文档（如pdf），相比于简单的OCR，多模态能够额外提供
- 文档布局
- 形状、图案、颜色、空间关系
- 扭曲文字的识别
等信息。

多模态能力的用途包括但不限于：
- 分析图片，视频，pdf中的特征：通过分析采样数据（如抽取视频的某一帧，从文件夹中选择一个pdf/图片文件），形成数据的大致轮廓，精准制定规划/确定下一步的方向。
- 将其他格式的文件渲染为图片，进行可视化分析
- debug/检查/验证/查漏补缺：对使用常规手段的批处理结果中的可疑数据或者抽样数据，使用多模态能力进行二次检验。

如果你能够调用外部LLM的话，你还可以使用外部LLM的多模态能力。

# 自身和外部LLM的文本/多模态能力的使用原则

## 禁用自身和外部LLM的文本/多模态能力的情形
### 需要准确的颜色/文本等信息时
你自身的多模态能力，以及外部LLM的文本/多模态能力，有时可能存在幻觉，并不准确，仅供参考。
当需要非常准确的信息时，应该使用专用的命令/SDK/模型（当某个命令/SDK模型不可用时，不要放弃，可以尝试以下同类型的其它替代品)。
### 数据量过大时
当数据量过大（如文本大于500k）或者文件过多（大于30个）时，禁止使用自身或者外部LLM的文本/多模态能力来处理数据，因为**时间、金钱成本会非常高昂**。
这时你必须使用其它文本或音视频处理手段（如OCR等)，但可以使用多模态能力来对采样结果进行分析，或者对部分处理结果进行验证。

## 其它情况下应优先使用自身多模态能力（相比于调用外部LLM）：
仅在以下情形考虑调用外部LLM的多模态能力：
- 你自身没有多模态能力
- 需要处理的数据量过大（如文本大于100k）或者文件过多（大于10个）
- 需要外部LLM的结果作为参考或者交叉验证
因为相比于调用外部LLM，使用你自身的多模态能力有以下优势：
- 保留更多细节信息，避免外部LLM转述造成信息丢失
- 减少2轮交互，节省时间和金钱成本，提高效率

# 效率
任务的处理效率非常重要，因此你必须充分使用缓存、批处理等技巧来提高效率。

## 高耗时环节的处理原则
当运行的某个命令，或者代码中的某个步骤，由于计算量较大或者需要访问网络等原因，耗时可能会很高时，必须遵循以下原则：

### 及时保存中间处理结果
为了避免重复执行耗时环节，应及时保存中间处理结果。典型场景包括：
- 批处理中的部分数据处理出错，导致整个批次需要重跑
- 后续处理环节执行失败，需要重新运行前面的环节
- 需要利用中间结果进行验证、调试或分析
通过保存中间结果，可以避免重复计算，提高效率。

### 周期性输出处理进度
对于耗时较长的环节，应定期输出处理进度（最好带上时间戳），以便
- 了解当前进展
- 预估整体任务的剩余执行时间
- 及时发现卡顿、死锁等异常情况
同时注意，处理进度的输出不应该过于频繁，以免产生过多无用信息，导致关键信息被淹没。
可以选择间隔一定时间（如每隔10秒）或者处理完一定量的数据（如每处理30个文件）后输出一次进度。

### 输出/打印处理进度、日志、数据等关键信息时，务必及时flush
如果输出时不进行flush操作，输出内容可能会一直在进程内部缓存，直到进程正常退出后才能获取到。
对于运行时间较长的命令/代码，这会导致无法及时看到相关输出，严重影响开发、诊断和调试效率。

## 必要时可以手动处理corner case
一些常规方法，如：
- 基于关键字或者正则表达式的文本处理方法
- 基于小模型的文本/图片/视频处理方法
经常会遇到难以处理的corner case，当corner case数据量较小时，使用多模态能力手动处理/修正，相比不断优化代码、训练模型花费的成本（时间、金钱）更低。


# 你的几个阶段
## 执行阶段
- 这是你的初始阶段，在这个阶段你使用工具来尝试执行todo。
- 当你认为执行阶段可以结束时（比如当前todo完成了，或者已经无法取得进展了，或者你想建议planner修改规划和todo时）时，你需要调用{END_EXECUTION_TOOL_NAME}工具。
- 该工具被调用后然后会进入汇报阶段。 
- **你只能在执行阶段调用工具，在之后的阶段（如汇报和QA）里你不能调用任何工具。**
每个executor最多可以执行{EXECUTOR_MAX_STEPS}步，超过限制后系统会剥夺其工具执行权，立即进入汇报阶段。

## 汇报阶段
在汇报阶段，{EXECUTOR_FEEDBACK_INSTRUCTION}。

## QA阶段
汇报阶段结束后，你会进入QA阶段，你会收到来自其他人的问题。

{TASK_HISTORY_INSTRUCTION}

{AGENT_COMMON_INSTRUCTION}
"""

# 注意：**此阶段可以和执行阶段合并在一起，即你可以在调用{END_EXECUTION_TOOL_NAME}工具的同时，\
# 直接将你的汇报内容作为常规的文本消息返回(注意不是放在tool call里，而是和tool_call并列），这样可以减少一次交互，提高效率。**

logger = getLogger(__name__)


class ExecutorState(Enum):
    READY = 0
    WAITING_TOOL_RESULT = 1


def get_ask_targets(task_history: TaskHistory) -> AskTargets:
    ask_targets: AskTargets = {}
    for member_result in task_history.team_member_results:
        member_id = member_result.member_id
        ask_targets[member_id.name] = member_id

    return ask_targets


class Executor(QAAgent, ReactAgent):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._state = ExecutorState.READY

        self._execution_end_reason: ExecutionEndReason | None = None
        self._extra_finished_todo_indexes: list[int] = []

    def _react_exit_prompt(self) -> str:
        return EXECUTOR_NOT_ISSUE_ANY_TOOL_CALL_INSTRUCTION

    @message_handler()
    async def start_execute(self, subtask: Subtask, ctx: MessageContext) -> None:
        self._sub_task = subtask
        self._step_count = 0
        self._tool_call_count = 0

        ask_member_tool = AskMemberTool(ctx)
        ask_member_tool.set_ask_targets(get_ask_targets(self._sub_task.task_history))
        self._internal_tool_registry = ToolRegistry(
            [
                Tool.from_callable(ask_member_tool, ASK_MEMBER_TOOL_NAME),
                Tool.from_callable(self.end_execution, END_EXECUTION_TOOL_NAME),
            ],
        )
        self._tools = [
            tool_def.to_openai_format()
            for tool_def in chain(
                self._model_config.tools_definition,
                self._internal_tool_registry.get_tools_definition(),
            )
        ]
        # if self._need_explicit_cache() and self._tools:
        #     self._tools[-1]["cache_control"] = {"type": "ephemeral"}

        personal_dir = self._id.name
        await self._publish_output_event(CreateAgentFolder(folder_name=personal_dir))

        prompt = "\n".join(
            [
                subtask.task.to_prompt(),
                self._task_config.to_prompt(personal_dir),
                "为了避免多段文本和格式造成混淆，部分内容用xml格式表示。",
                "# 任务执行历史",
                subtask.task_history.to_prompt(),
                "",
                "# planner的当前规划",
                subtask.plan.to_prompt(),
                "",
                f"你的名字是`{self._id.name}`，你现在负责规划中的第{subtask.todo_index + 1}项，即：{subtask.plan.todos[subtask.todo_index].desc}",
            ]
        )

        executor_instruction = EXECUTOR_INSTRUCTION
        if subtask.task.env_notes:
            executor_instruction += subtask.task.env_notes

        system_part: dict = {
            "type": "text",
            "text": executor_instruction,
        }
        if self._need_explicit_cache():
            system_part["cache_control"] = {"type": "ephemeral"}
        self._messages = [
            {
                "role": "system",
                "content": [system_part],
            },  # pyright: ignore[reportAttributeAccessIssue]
            {"role": "user", "content": prompt},
        ]

        await self._publish_model_input(self._messages)
        await self._do_turn(ctx)

    async def _on_turn_done(self, ctx: MessageContext) -> None:
        # TODO：限制最大token数
        if self._execution_end_reason is None:
            self._step_count += 1
            if self._step_count < EXECUTOR_MAX_STEPS:
                await self._do_turn(ctx)
            else:
                self._execution_end_reason = ExecutionEndReason.STEPS_EXHAUSTED
                report_content = await self._gen_report(True, ctx)
                await self._send_report(report_content, ctx)
        else:
            # if self._tool_message.content:
            #     report_content = self._tool_message.content
            # else:
            report_content = await self._gen_report(False, ctx)
            await self._send_report(report_content, ctx)

    async def _gen_report(self, steps_exhausted: bool, ctx: MessageContext) -> str:
        if steps_exhausted:
            feed_back_instruction = EXECUTOR_STEPS_EXAHUSTED_INSTRUCTION
            self._messages.append(
                {
                    "role": "user",
                    "content": feed_back_instruction,
                }
            )
            cache_control_injection_points = None
        else:
            cache_control_injection_points = [
                {
                    "location": "message",
                    "index": -3,
                }
            ]

        choice = await self._acompletion_without_tool_calls(
            messages=self._messages,
            tools=self._tools,
            cache_control_injection_points=cache_control_injection_points,
        )
        self._messages.append(choice.message)
        assert choice.message.content
        return choice.message.content

    async def _send_report(self, report_content: str, ctx: MessageContext):
        from judy_core.agents.planner import Planner

        assert self._execution_end_reason is not None

        await ctx.send(
            ExecutorResult(
                member_id=TeamMemberId.from_actor_id(self._id),
                end_reason=self._execution_end_reason,
                plan=self._sub_task.plan,
                todo_index=self._sub_task.todo_index,
                extra_finished_todo_indexes=self._extra_finished_todo_indexes,
                report_content=report_content,
                step_count=self._step_count,
                tool_call_count=self._tool_call_count,
            ),
            ActorId.default(Planner.actor_kind),
        )

    async def end_execution(
        self,
        reason: ExecutionEndReason,
        extra_finished_todo_indexes: Annotated[
            list[int] | None,
            Field(
                description="如果你本次执行期间不仅完成了当前todo，还完成了其他todo，可以把他们的序号加到这里"
            ),
        ] = None,
    ):
        """当你认为你可以停止执行的时候（即当你负责的todo已经完成，或者失败，或者你想向planner提出建议时），调用这个工具。

        该工具被调用后会立刻进入汇报阶段，因此你只能在最后一步才能调用这个工具。否则你将被提前中止执行，立即进入汇报阶段。
        """
        self._execution_end_reason = reason

        self._extra_finished_todo_indexes = []
        if (
            reason == ExecutionEndReason.COMPLETED
            and extra_finished_todo_indexes is not None
        ):
            for extra_finished_todo_index in extra_finished_todo_indexes:
                extra_finished_todo_index -= 1
                if extra_finished_todo_index == self._sub_task.todo_index:
                    continue
                if extra_finished_todo_index < 0:
                    continue
                if extra_finished_todo_index >= len(self._sub_task.plan.todos):
                    continue
                self._extra_finished_todo_indexes.append(extra_finished_todo_index)

        return TOOL_RESULT_SUCCESS_CONTENT + ". " + EXECUTOR_FEEDBACK_PHASE_INSTRUCTION



import json
import traceback
from collections.abc import AsyncGenerator
from copy import deepcopy
from datetime import datetime
from typing import Annotated, Any, cast

import contexttimer
from judy_core.utils.llm import acompletion
from litellm import Choices
from litellm.types.llms.vertex_ai import VertexToolName
from litellm.types.utils import ModelResponse
from pydantic import BaseModel, Field

CALL_LLM_BATCH_TOOL_NAME = "call_llm_batch"

CallLLMId = str | None

SYSTEM_INSTRUCTIOINS_FIELD_DESCRIPTION = """\
不需要填写role字段，直接填写content内容即可。
样例1：
"..."
样例2：
[
    {"type":"text", "text":"..."}, 
    {"type":"image_url", "image_url":{"url":"data:image/jpeg;base64,..."}}
]
\
"""

MESSAGES_FIELD_DESCRIPTION = """\
与litellm的message格式相同（与openai格式类似），至少需要包含role,content字段。样例：
[
    {
        "role": "user",
        "content": [
            {"type":"text", "text":"..."}, 
            {"type":"image_url", "image_url":{"url":"data:image/jpeg;base64,..."}}
        ]
    },
    ...
]\
"""


class CallLLMInput(BaseModel):
    id: CallLLMId

    system_instructions: Annotated[
        str | list[dict], Field(description=SYSTEM_INSTRUCTIOINS_FIELD_DESCRIPTION)
    ]
    messages: Annotated[
        list[dict],
        Field(description=MESSAGES_FIELD_DESCRIPTION),
    ]

    response_json_schema: Annotated[
        dict[str, Any] | None, Field(description="设置structured output的json schema")
    ] = None
    tool_names: list[str] | None = None


class CallLLMBatchInput(BaseModel):
    inputs: list[CallLLMInput]


class CallLLMOutput(BaseModel):
    error: Annotated[str | None, Field(description="错误信息")]
    id: CallLLMId
    text: str


class CallLLMOutputWithDebugInfo(BaseModel):
    output: CallLLMOutput
    time_cost: float
    timestamp: str
    raw_errmsg: str | None = None
    response: ModelResponse | None = None


class CallLLMBatchOutput(BaseModel):
    outputs: list[CallLLMOutput]


async def call_llm_batch(input_json_file_path: str, output_json_file_path: str) -> str:
    """你可以使用此工具来处理复杂的文本任务（如分类、信息提取等）

    input_json_file_path: json schema如下：{input_json_schema}
    output_json_file_path: json schema 如下： {output_json_schema}
    """
    ...


if call_llm_batch.__doc__ is not None:
    call_llm_batch.__doc__ = call_llm_batch.__doc__.format(
        input_json_schema=json.dumps(CallLLMBatchInput.model_json_schema()),
        output_json_schema=json.dumps(CallLLMBatchOutput.model_json_schema()),
    )

CLAUDE_TOOL_MAX_USE = 5
CLAUDE_WEB_SEARCH_TOOL_NAME = "web_search"
CLAUDE_WEB_SEARCH_TOOL_TYPE = "web_search_20250305"
CLAUDE_WEB_FETCH_TOOL_NAME = "web_fetch"
CLAUDE_WEB_FETCH_TOOL_TYPE = "web_fetch_20250910"
CLAUDE_TOOL_DICT: dict[str, dict] = {
    CLAUDE_WEB_SEARCH_TOOL_NAME: {
        "type": CLAUDE_WEB_SEARCH_TOOL_TYPE,
    },
    CLAUDE_WEB_FETCH_TOOL_NAME: {
        "type": CLAUDE_WEB_FETCH_TOOL_TYPE,
    },
}

WEB_FETCH_TOOL_DESCRIPTION = "网页抓取，访问URL并提取网页或文件内容。对于高度动态和比较复杂的网页，抓取工具能够提供更准确的内容。"


class LLMBuiltinTool(BaseModel):
    name: str
    description: str


def get_llm_builtin_tools_info(model_name: str) -> list[LLMBuiltinTool]:
    if "gemini" in model_name:
        return [
            LLMBuiltinTool(
                name=VertexToolName.GOOGLE_SEARCH.value,
                description="google网页搜索，联网搜索最新信息。不支持指定搜索的时间范围和网页快照，只能搜索到网页的当前最新信息。如果查询的是过去某个时间点的历史信息，该工具给出的结果并不可靠。",
            ),
            LLMBuiltinTool(
                name=VertexToolName.URL_CONTEXT.value,
                description=WEB_FETCH_TOOL_DESCRIPTION,
            ),
        ]
    elif "claude" in model_name:
        return [
            LLMBuiltinTool(
                name=CLAUDE_WEB_SEARCH_TOOL_NAME,
                description="网页搜索，联网搜索最新信息",
            ),
            # LLMBuiltinTool(
            #     name=CLAUDE_WEB_FETCH_TOOL_NAME,
            #     description=WEB_FETCH_TOOL_DESCRIPTION,
            # ),
        ]

    else:
        return []


def get_llm_builin_tools(model_name: str, tool_names: list[str] | None) -> list | None:
    if not tool_names:
        return None
    tools = []
    if "gemini" in model_name:
        for tool_name in tool_names:
            tools.append({tool_name: {}})
    elif "claude" in model_name:
        for tool_name in tool_names:
            tool = CLAUDE_TOOL_DICT.get(tool_name)
            if tool is None:
                raise RuntimeError("unsupported tool: {}".format(tool_name))
            tool = deepcopy(tool)
            tool["name"] = tool_name
            tool["max_use"] = CLAUDE_TOOL_MAX_USE
            tools.append(tool)
    else:
        raise RuntimeError("unsupported tool: {}".format(tool_names))
    return tools


async def call_llm_batch_impl(
    model_name: str, batch_input: CallLLMBatchInput
) -> AsyncGenerator[CallLLMOutputWithDebugInfo]:
    for i, input in enumerate(batch_input.inputs):
        raw_errmsg = None
        with contexttimer.Timer() as t:
            try:
                response_format = None
                if input.response_json_schema is not None:
                    response_format = {
                        "type": "json_schema",
                        "json_schema": {
                            "name": "output_schema",
                            "schema": input.response_json_schema,
                            "strict": True,
                        },
                    }
                tools = get_llm_builin_tools(model_name, input.tool_names)

                system_message = {
                    "role": "system",
                    "content": input.system_instructions,
                }
                response = await acompletion(
                    model=model_name,
                    messages=[system_message] + input.messages,
                    response_format=response_format,
                    tools=tools,
                    timeout=90,
                )
                if response is not None:
                    if len(response.choices) > 0:
                        choice = cast(Choices, response.choices[0])
                        output = CallLLMOutput(
                            error=None,
                            id=input.id,
                            text=choice.message.content or "",
                        )

                    else:
                        output = CallLLMOutput(
                            error="model call return a response with no choices",
                            id=input.id,
                            text="",
                        )
                else:
                    output = CallLLMOutput(
                        error="model call return a null response",
                        id=input.id,
                        text="",
                    )
            except Exception as e:
                response = None
                raw_errmsg = "".join(traceback.format_exception(e, limit=3))
                truncated_tb_msg = "".join(raw_errmsg.splitlines()[-10:])
                output = CallLLMOutput(
                    error="model call failed, error：{}".format(truncated_tb_msg),
                    id=input.id,
                    text="",
                )
        yield CallLLMOutputWithDebugInfo(
            output=output,
            raw_errmsg=raw_errmsg,
            time_cost=t.elapsed,
            timestamp=str(datetime.now()),
            response=response,
        )


def summary_batch_output(
    batch_output: CallLLMBatchOutput, output_json_file_path: str
) -> str:
    count = len(batch_output.outputs)
    error_count = 0
    for output in batch_output.outputs:
        if output.error is not None:
            error_count += 1
    if error_count != 0:
        return "Some input process failed, error_count:{}, success_count:{}, process result(including failed) have been saved to {}.".format(
            error_count, count - error_count, output_json_file_path
        )
    return "All inputs were processed successfully."



from judy_core.types import InlineBlobPart, TextPart
from judy_core.utils.tool import ContentUnion
from litellm.types.completion import (
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartTextParam,
    ImageURL,
)
from litellm.types.llms.anthropic import (
    AnthropicContentParamSource,
    AnthropicMessagesDocumentParam,
)


def convert_blob_to_image_part(
    blob: InlineBlobPart,
) -> ChatCompletionContentPartImageParam:
    image_url = "data:{};{},{}".format(blob.mime_type, blob.encode_type, blob.data)
    image_part = ChatCompletionContentPartImageParam(
        type="image_url",
        image_url=ImageURL(url=image_url),
    )
    return image_part


def convert_blob_to_document_part(
    blob: InlineBlobPart,
) -> AnthropicMessagesDocumentParam:
    doc_part = AnthropicMessagesDocumentParam(
        type="document",
        source=AnthropicContentParamSource(
            type="base64", media_type=blob.mime_type, data=blob.data
        ),
    )
    return doc_part


def convert_blob_to_part(
    blob: InlineBlobPart, model: str | None = None
) -> ChatCompletionContentPartImageParam | AnthropicMessagesDocumentParam:
    if model:
        if "claude" in model:
            if blob.mime_type.startswith("image/"):
                return convert_blob_to_image_part(blob)
            else:
                return convert_blob_to_document_part(blob)
    return convert_blob_to_image_part(blob)


def convert_content_to_openai_format(
    content: ContentUnion,
    model: str | None = None,
) -> str | list[ChatCompletionContentPartParam]:
    if isinstance(content, str):
        return content
    else:
        message_content = []
        for part in content.parts:
            if isinstance(part, TextPart):
                message_content.append(
                    ChatCompletionContentPartTextParam(
                        {"type": "text", "text": part.text}
                    )
                )
            else:
                blob: InlineBlobPart = part
                part = convert_blob_to_part(blob, model)
                message_content.append(part)
        return message_content



import asyncio
import time
import traceback
from copy import deepcopy
from typing import Protocol, cast

import httpx
import litellm
from judy_core.configs import JudyCompletionOption
from litellm import Choices, CustomStreamWrapper
from litellm.exceptions import (
    InternalServerError,
    RateLimitError,
    ServiceUnavailableError,
)
from litellm.exceptions import Timeout as LitellmTimeout
from litellm.llms.vertex_ai.common_utils import _build_vertex_schema
from litellm.types.utils import Message, ModelResponse, ModelResponseStream


class AsyncLogger(Protocol):
    async def __call__(self, text: str) -> None: ...


BLANK_RESPONSE_RETRY_DEFAULT_MESSAGE = """Ignore this message. This message is only intended to trigger your reply. \
You should give a reasonable response based on the current conversation history, as if this message did not exist at all.
"""


def _remove_warnings_from_litellm_completion(messages):
    # 升级litellm版本后，出现大量如下类似警告，暂时用这种方式抑制，待litellmt解决这个问题后，这部分代码可以取出
    # pydantic/main.py:464:UserWarning: Pydantic serializer warnings:
    # PydanticSerializationUnexpectedValue(Expected `ChatCompletionMessageToolCall` - serialized value may not be as expected [field_name='tool_calls', input_value={'index': 0, 'provider_sp...+4', 'type': 'function'}, input_type=dict])
    new_messages = []
    for message in messages:
        if isinstance(message, Message):
            message = message.model_dump()
        new_messages.append(message)
    return new_messages


def adapt_tools_for_model(model: str, tools: list[dict] | None) -> list[dict] | None:
    if not tools:
        return tools

    if not ("openrouter" in model and "gemini" in model):
        return tools

    adapted_tools = []
    for tool in tools:
        adapted_tool = deepcopy(tool)
        function = adapted_tool.get("function")
        if function:
            parameters = function.get("parameters")
            if parameters:
                adapted_parameters = _build_vertex_schema(parameters)
                function["parameters"] = adapted_parameters
        adapted_tools.append(adapted_tool)
    return adapted_tools


def enable_claude_opus_thinking(args: dict) -> None:
    if "thinking" not in args:
        args["thinking"] = {"type": "adaptive"}
    allowed_openai_params = args.setdefault("allowed_openai_params", [])
    allowed_openai_params.append("thinking")


def adapt_completion_params(model: str, **kwargs) -> dict:
    if "gemini" in model:
        if "reasoning_effort" not in kwargs:
            kwargs["reasoning_effort"] = "high"
        # if "temperature" not in kwargs:
        #     kwargs["temperature"] = 1.0
        # kwargs["toolConfig"] = {
        #     "functionCallingConfig": {
        #         "stream_function_call_arguments": True,
        #     }
        # }
    elif "opus-4.6" in model or "opus-4-6" in model:
        if "max_tokens" not in kwargs:
            kwargs["max_tokens"] = 128 * 1024
        if "effort" not in kwargs:
            kwargs["effort"] = "max"
        if model.startswith("anthropic/"):
            enable_claude_opus_thinking(kwargs)
        else:
            extra_body = kwargs.setdefault("extra_body", {})
            enable_claude_opus_thinking(extra_body)
    return kwargs


class StreamChunkTimeout(Exception):
    def __init__(self, msg: str, ex: Exception):
        super().__init__(msg)
        self.ex = ex


JUDY_HIDDEN_PARAMS_KEY = "judy"
RESPONSE_TIME_COST = "response_time_cost"
FIRST_TOKEN_TIME_COST = "first_token_time_cost"
STREAM_TIME_COST = "stream_time_cost"
STREAM_CHUNKS_LEN = "stream_chunks_len"


async def _acompletion(
    model: str,
    *,
    messages,
    judy_option: JudyCompletionOption,
    stream: bool | None = None,
    stream_options: dict | None = None,
    **kwargs,
) -> ModelResponse:
    if stream:
        default_stream_options = {"include_usage": True}
        if stream_options is not None:
            default_stream_options.update(stream_options)
        stream_options = default_stream_options
    start_time = time.perf_counter()
    response = await litellm.acompletion(
        model,
        messages=messages,
        stream=stream,
        stream_options=stream_options,
        **kwargs,
    )
    response_time = time.perf_counter()
    response_time_cost = response_time - start_time
    judy_hidden_params = {
        RESPONSE_TIME_COST: response_time_cost,
    }
    if isinstance(response, CustomStreamWrapper):
        first_token_time = None
        chunks = []
        response_iter = aiter(response)
        while True:
            wait_start_time = time.perf_counter()
            try:
                async with asyncio.timeout(judy_option.stream_chunk_timeout):
                    chunk = await anext(response_iter)
                    if first_token_time is None:
                        first_token_time = time.perf_counter()
                        judy_hidden_params[FIRST_TOKEN_TIME_COST] = (
                            first_token_time - response_time
                        )
            except StopAsyncIteration:
                break
            except Exception as e:
                if not isinstance(e, httpx.ReadTimeout | TimeoutError):
                    raise
                current_time = time.perf_counter()
                errmsg = "stream chunk timeout, response_id: {}, response_time_cost:{}, stream_time_cost: {}, current_chunk_wait_time_cost:{}, stream_chunks: {}".format(
                    response.response_id,
                    response_time_cost,
                    current_time - response_time,
                    current_time - wait_start_time,
                    len(chunks),
                )
                raise StreamChunkTimeout(errmsg, e)
            chunk = cast(ModelResponseStream, chunk)
            chunks.append(chunk)
        stream_end_time = time.perf_counter()
        judy_hidden_params[STREAM_TIME_COST] = stream_end_time - response_time
        judy_hidden_params[STREAM_CHUNKS_LEN] = len(chunks)
        # chunks[-1] = response.chunks[-1]
        model_response = litellm.stream_chunk_builder(chunks, messages)
        assert isinstance(model_response, ModelResponse)
    else:
        model_response = cast(ModelResponse, response)

    if getattr(model_response, "_hidden_params", None) is None:
        model_response._hidden_params = {}
    model_response._hidden_params[JUDY_HIDDEN_PARAMS_KEY] = judy_hidden_params
    return model_response


async def _acompletion_with_retries(
    model: str,
    *,
    judy_option: JudyCompletionOption | None = None,
    judy_logger: AsyncLogger | None = None,
    **kwargs,
) -> ModelResponse:
    retry_count = 0
    if judy_option is None:
        judy_option = JudyCompletionOption()

    while True:
        start_time = time.perf_counter()
        try:
            model_response = await _acompletion(
                model,
                judy_option=judy_option,
                **kwargs,
            )
            return model_response
        except RateLimitError as e:
            ex = e
            errmsg = "rate limit"
        except LitellmTimeout as e:
            ex = e
            errmsg = "response timeout"
        except StreamChunkTimeout as e:
            ex = e.ex
            errmsg = str(e)
        except Exception as e:
            if not isinstance(
                e, ServiceUnavailableError | InternalServerError | httpx.ReadTimeout
            ):
                raise
            ex = e
            errmsg = "unknown exception"

        if judy_logger:
            time_cost = time.perf_counter() - start_time
            await judy_logger(
                "completion error, total_time_cost: {}, errmsg: {}, type:{}, ex: {}".format(
                    time_cost,
                    errmsg,
                    type(ex).__name__,
                    "\n".join(traceback.format_exception(ex)),
                )
            )

        if retry_count < judy_option.num_retries:
            retry_count += 1
        else:
            raise ex


async def acompletion(
    model: str,
    *,
    messages,
    tools: list[dict] | None = None,
    judy_option: JudyCompletionOption | None = None,
    judy_logger: AsyncLogger | None = None,
    blank_response_retry_message: str | None = None,
    **kwargs,
) -> ModelResponse | None:
    api_call_count = 0
    while True:
        try:
            adapted_tools = adapt_tools_for_model(model, tools)
            adapted_kwargs = adapt_completion_params(model, **kwargs)
            response = await _acompletion_with_retries(
                model,
                messages=_remove_warnings_from_litellm_completion(messages),
                tools=adapted_tools,
                judy_option=judy_option,
                judy_logger=judy_logger,
                stream=True,
                **adapted_kwargs,
            )
            response = cast(ModelResponse, response)
        except InternalServerError as e:
            if "Invalid response object" not in e.message:
                raise
            else:
                response = ModelResponse(choices=[])

        if len(response.choices) == 0:
            # gemini api有时返回内容为空（没有candidates)，出现这种情况后尝试重新生成
            # https://discuss.ai.google.dev/t/gemini-2-5-pro-with-empty-response-text/81175/356
            if judy_logger:
                await judy_logger(f"response has no choice: {response}")
            api_call_count += 1
            if api_call_count < 3:
                continue
            elif api_call_count == 3:
                if blank_response_retry_message is None:
                    blank_response_retry_message = BLANK_RESPONSE_RETRY_DEFAULT_MESSAGE
                messages.append(
                    {"role": "user", "content": blank_response_retry_message}
                )
            elif api_call_count < 5:
                continue
            else:
                return None
                # raise ModelBlankResponseError(
                #     f"No choices returned from LLM: {response}"
                # )
        else:
            break

    return response


async def acompletion_choice(**kwargs) -> Choices | None:
    response = await acompletion(**kwargs)
    if response is None:
        return None
    choice = cast(Choices, response.choices[0])
    return choice



import xml.etree.ElementTree as ET
from datetime import datetime
from enum import StrEnum
from typing import Annotated

from pydantic import BaseModel, Field
from tongcell.core import ActorId

from judy_core.configs import JudyCompletionOption, LitellmCompletionOption
from judy_core.utils.common import truncate_text
from judy_core.utils.tool import ToolDefinition

TASK_PROMPT_TEMPLATE = """\
# 任务
<task>
{task_desc}
</task>
"""


class Task(BaseModel):
    task_desc: str
    env_notes: str | None = None
    verifier_mandatory_checklist: dict[str, str] | None = None

    def to_prompt(self) -> str:
        content = TASK_PROMPT_TEMPLATE.format(task_desc=self.task_desc)
        return content


class TodoStatus(StrEnum):
    PENDING = "PENDING"
    # IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"


TeamMemberName = str


class TeamMemberId(BaseModel):
    kind: str
    name: TeamMemberName

    @classmethod
    def from_actor_id(cls, actor_id: ActorId):
        return cls(kind=actor_id.kind, name=actor_id.name)


class Todo(BaseModel):
    status: TodoStatus
    desc: str

    def to_xml_element(self) -> ET.Element:
        elem = ET.Element("todo")
        elem.text = self.desc
        return elem


class Plan(BaseModel):
    task_finished: Annotated[
        bool, Field(description="任务完成标记。在无需进一步规划时设置为为true。")
    ]
    todos: list[Todo]

    def to_prompt(self) -> str:
        plan_elem = ET.Element("plan")
        for i, todo in enumerate(self.todos):
            todo_element = todo.to_xml_element()
            todo_element.attrib["index"] = str(i + 1)
            todo_element.attrib["status"] = str(todo.status)
            plan_elem.append(todo_element)

        ET.indent(plan_elem, space=" ", level=0)
        return ET.tostring(plan_elem, encoding="unicode")

    def str_for_human(self) -> str:
        segs = [f"task_finished: {self.task_finished}"]
        for i, todo in enumerate(self.todos):
            todo_desc_lines = todo.desc.split("\n")
            todo_desc = "\n   ".join(todo_desc_lines)
            segs.append(f"{i + 1}. [{todo.status}] {todo_desc}")
        return "\n".join(segs)


class ExecutionEndReason(StrEnum):
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    HAVE_SUGGESTIONS = "HAVE_SUGGESTIONS"
    STEPS_EXHAUSTED = "STEPS_EXHAUSTED"

    def str_for_model(self) -> str:
        if self == ExecutionEndReason.COMPLETED:
            return "成功"
        elif self == ExecutionEndReason.FAILED:
            return "失败"
        elif self == ExecutionEndReason.HAVE_SUGGESTIONS:
            return "提建议"
        elif self == ExecutionEndReason.STEPS_EXHAUSTED:
            return "步数耗尽"
        return "未知"


class ExecutorResult(BaseModel):
    member_id: TeamMemberId
    end_reason: ExecutionEndReason

    plan: Plan
    todo_index: int
    extra_finished_todo_indexes: list[int]
    report_content: str

    step_count: int
    tool_call_count: int


class VerificationRequest(BaseModel):
    task: Task
    task_execution_summary: str


class VerificationResultStatus(StrEnum):
    PASSED = "PASSED"
    FAILED = "FAILED"
    STEPS_EXHAUSTED = "STEPS_EXHAUSTED"


class VerificationResult(BaseModel):
    member_id: TeamMemberId
    status: VerificationResultStatus
    report_content: str


def executor_result_for_model(executor_result: ExecutorResult) -> ET.Element:
    executor_result_elem = ET.Element("executor_result")
    executor_result_elem.attrib["member_name"] = executor_result.member_id.name
    executor_result_elem.attrib["end_reason"] = executor_result.end_reason
    executor_result_elem.attrib["step_count"] = str(executor_result.step_count)
    executor_result_elem.attrib["tool_call_count"] = str(
        executor_result.tool_call_count
    )

    todo = executor_result.plan.todos[executor_result.todo_index]
    todo_element = todo.to_xml_element()
    executor_result_elem.append(todo_element)
    for index in executor_result.extra_finished_todo_indexes:
        extra_todo = executor_result.plan.todos[index]
        todo_element = extra_todo.to_xml_element()
        executor_result_elem.append(todo_element)

    REPORT_CONTENT_MAX_LEN = 2000
    report_truncation_result = truncate_text(
        executor_result.report_content, REPORT_CONTENT_MAX_LEN
    )
    report_elem = ET.Element("report_content")
    report_elem.attrib["length"] = str(len(executor_result.report_content))
    report_elem.text = report_truncation_result.text

    executor_result_elem.append(report_elem)
    return executor_result_elem


def verification_result_for_model(
    verification_result: VerificationResult,
) -> ET.Element:
    elem = ET.Element("verification_result")
    elem.attrib["member_name"] = verification_result.member_id.name
    elem.attrib["status"] = verification_result.status
    elem.text = verification_result.report_content
    return elem


type TeamMemberResult = ExecutorResult | VerificationResult


class TaskHistory(BaseModel):
    team_member_results: list[TeamMemberResult]

    def to_prompt(self):
        task_history_elem = ET.Element("task_history")
        for team_member_result in self.team_member_results:
            if isinstance(team_member_result, ExecutorResult):
                elem = executor_result_for_model(team_member_result)
            else:
                elem = verification_result_for_model(team_member_result)
            task_history_elem.append(elem)
        ET.indent(task_history_elem, space=" ", level=0)
        return ET.tostring(task_history_elem, encoding="unicode")


class Subtask(BaseModel):
    task: Task
    task_history: TaskHistory
    plan: Plan
    todo_index: int


BASIC_ENV_INFO_PROMPT_TEMPLATE = """\
# 一些可能有用的环境信息
这是系统检测到的一些环境信息，并不全面，仅供参考，更详细的需要自行探索。
<basic_env_info>
{basic_env_info}
</basic_env_info>
"""


class TaskConfig(BaseModel):
    space_dir: str
    task_dir: str | None = None
    basic_env_info: str | None = None
    disable_verifier: bool = False

    def to_prompt(self, agent_dir: str | None = None) -> str:
        segs = ["# 目录"]
        if self.task_dir:
            segs.append(
                "任务目录（默认工作目录）: `{task_dir}`".format(task_dir=self.task_dir)
            )
        segs.append("团队空间目录: `{space_dir}`".format(space_dir=self.space_dir))
        if agent_dir:
            segs.append(
                "你的个人目录: `{space_dir}/{agent_dir}`".format(
                    space_dir=self.space_dir, agent_dir=agent_dir
                )
            )

        segs.append("")

        if self.basic_env_info:
            segs.append(
                BASIC_ENV_INFO_PROMPT_TEMPLATE.format(
                    basic_env_info=self.basic_env_info
                )
            )
        content = "\n".join(segs)
        return content


class GetTaskConfig(BaseModel):
    pass


class GetModelConfig(BaseModel):
    pass


class ModelConfig(BaseModel):
    model: str
    tools_definition: list[ToolDefinition]
    judy_completion_option: JudyCompletionOption = Field(
        default_factory=JudyCompletionOption
    )
    litellm_completion_option: LitellmCompletionOption = Field(
        default_factory=LitellmCompletionOption
    )


class MemberQuestion(BaseModel):
    text: str


class MemberAnswer(BaseModel):
    text: str


class GetTaskState(BaseModel):
    pass


class TaskState(BaseModel):
    task_start_ts: float

    def __init__(self):
        super().__init__(task_start_ts=datetime.now().timestamp())

    def elapsed_time_sec(self) -> float:
        return datetime.now().timestamp() - self.task_start_ts


class ShellCommandCompletedNotification(BaseModel):
    text: str


class GetEnvironmentNotification(BaseModel):
    pass


type EnvironmentNotification = ShellCommandCompletedNotification


class EnvironmentNotificationList(BaseModel):
    notifications: Annotated[list[EnvironmentNotification], Field()] = []



from judy_core.messages import (
    MemberAnswer,
    MemberQuestion,
    TeamMemberId,
    TeamMemberName,
)
from tongcell.core import ActorId, ActorKind, MessageContext

ASK_MEMBER_TOOL_NAME = "ask"

type AskTargets = dict[TeamMemberName, TeamMemberId]


class AskMemberTool:
    def __init__(self, ctx: MessageContext):
        self._ctx = ctx
        self._targets: AskTargets = {}

    def set_ask_targets(self, targets: AskTargets) -> None:
        self._targets = targets

    async def __call__(self, member_name: TeamMemberName, question_text: str):
        """向团队成员(executor, verifier)提问"""
        target_member_id = self._targets.get(member_name)
        if target_member_id is None:
            return "提问失败，成员不存在或者你无权向其提问: {}".format(member_name)

        target_actor_id = ActorId(
            ActorKind(target_member_id.kind), target_member_id.name
        )

        answer = await self._ctx.request(
            MemberQuestion(text=question_text),
            target_actor_id,
            response_type=MemberAnswer,
        )
        return answer.text



from judy_core.messages import GetModelConfig, GetTaskConfig, ModelConfig, TaskConfig
from tongcell.core import MessageContext, RoutedActor, message_handler


class Configuration(RoutedActor):
    def __init__(self, task_config: TaskConfig, model_config: ModelConfig):
        super().__init__()
        self._task_config = task_config
        self._model_config = model_config

    @message_handler()
    async def get_task_config(
        self, _: GetTaskConfig, ctx: MessageContext
    ) -> TaskConfig:
        return self._task_config

    @message_handler()
    async def get_configuration(
        self, _: GetModelConfig, ctx: MessageContext
    ) -> ModelConfig:
        from judy_core.agents.executor import Executor
        from judy_core.agents.verifier import Verifier

        model_config = self._model_config.model_copy()
        if ctx.envelope.sender.kind not in (Executor.actor_kind, Verifier.actor_kind):
            model_config.tools_definition = []
        return model_config



from itertools import chain

from judy_core.agents.prompts import AGENT_FOLDER_INSTRUCTION
from judy_core.agents.qa_agent import QAAgent
from judy_core.agents.react_agent import ReactAgent
from judy_core.events import CreateAgentFolder
from judy_core.messages import (
    Task,
    TeamMemberId,
    VerificationRequest,
    VerificationResult,
    VerificationResultStatus,
)
from judy_core.utils.tool import TOOL_RESULT_SUCCESS_CONTENT, Tool, ToolRegistry
from tongcell.core import ActorId, MessageContext, message_handler

FINISH_VERIFICATION_TOOL_NAME = "finish_verification"
VERIFIER_MAX_STEPS = 50

VERIFIER_STEPS_EXHAUSTED_PROMPT = """\
验证期间执行的步数已超限，中止验证。
"""

VERIFIER_FINISH_INSTRUCTION = f"""\
如果你认为已经可以出具验证结果和报告，你需要调用{FINISH_VERIFICATION_TOOL_NAME}工具来结束验证；否则你应该调用其它工具继续进行验证。\
"""

VERIFIER_REPORT_INSTRUCTION = """\
你需要就目前的发现和取得的进展进行汇报，汇报中可以包含你的建议。该汇报会被提交给planner，以便其及时调整规划。**你现在已经不能调用任何工具了，直接将汇报作为你的回复内容即可**。
"""

# TODO 添加构建测试用例的手段，如 黑盒测试 白盒测试
VERIFIER_MAIN_INSTRUCTION = f"""\
你(verifier)隶属于一个任务小组，该小组由一个planner、一个verifier和多个executor组成。
planner负责制定规划，executor负责执行规划中的各个todo，当planner认为任务已经完成后，会交给你(verifier)来验证/测试。
如果验证通过则任务完成，否则planner会根据你的反馈，调整规划，并分配executor来修正错误，修正之后会再次交给你来验证。

# 你的工作
你负责对根据任务要求，对交付的内容进行验证，并出具验证报告。
注意：
**禁止对交付的结果（代码或者数据文件）做任何修改。**
**不要尝试自行解决或者改进验证中发现的问题**。
你所发现的问题由planner和executor负责修复，他们有任务的完整历史信息和相关工具，在解决任务方面更有经验，更高效。


{AGENT_FOLDER_INSTRUCTION}
**作为verifier，交付目录对于你是只读的，禁止创建或者修改交付目录中的任何文件。你直接或者间接创建的文件（如编写的测试脚本、编译的产物等)，必须位于你的个人目录。**

# 测试case（验证case）
测试case来自于任务要求，因此在开始验证前，必须先分析任务中所提出的任务要求。注意：
- 当任务包含口语化表达或者非正式用语时，你必须敏锐地捕捉到其言外之意和隐性的要求。
- 当任务要求与环境冲突时，以任务要求为准。
详细内容见**任务要求**小结。
当所交付的解决方案不满足任务要求时，须判断为验收不通过。

## corner case
除非任务明确要求，你不必对非法输入等corner case进行验证，可以假定输入参数都是合法的参数。

### 测试脚本
如果任务提供了测试脚本，你必须运行该脚本来进行验证。
测试脚本可能需要特殊的运行方式才能得到正确的结果（比如用python解释器运行pytest脚本会导致进程正常退出，但实际上测试用例并没有被执行）。

### 测试数据集
有些问题可能是公开或者经典问题，这类问题通常会有公开的数据集。
对于此类问题，除了任务提供的数据集，你还可以通过下载额外的数据集来确保交付方案的覆盖率。

### 自定义验证方式
如果任务没有提供验证脚本，或者验证脚本的没有完全覆盖任务要求。那么你需要自行设计合理的验证方式。

### 交付结果的可重复性/幂等性
如果任务显式或者隐式要求（根据惯例或最佳实践）交付结果应该是可重复的/幂等的，应对交付结果的可重复性/幂等性进行验证。

# 你的两个阶段
## 验证阶段
在这个阶段，你需要根据任务要求，调用工具对交付的结果进行全面的验证。
{VERIFIER_FINISH_INSTRUCTION}

## 汇报阶段
在这个阶段，你需要就验证结果进行汇报，汇报内容会被提交给planner。

### 验证报告
当验证未通过时，为了帮助planner和executor定位问题和改进，你需要在验证报告中提供失败的case的复现方式（如验证脚本/数据的路径和运行方式）。

## 补充说明
一个任务可能会经历多次验证-修正循环，直至验证通过为止。因此你会在两个阶段间不断切换。

"""


# 你可以在调用{FINISH_VERIFICATION_TOOL_NAME}工具的同时，在tool call之外的回复内容中，同时给出你的汇报内容。
# 这样两个阶段就合并在了一起，可以减少一次交互延迟，提升效率。

CHECK_ITEM_ALL_TASK_REQUIREMENTS_AND_TEST_CASES_VERIFIED = """\
从任务描述中仔细分析出所有的：
- 需求
- 测试用例（应覆盖任务描述中提及的各类参数取值）
必须确保验证过程完整覆盖了以上内容。
"""

DEFAULT_VERIFIER_MANDATORY_CHECKLIST = {
    "all-task-requirements-and-test-cases-verified": CHECK_ITEM_ALL_TASK_REQUIREMENTS_AND_TEST_CASES_VERIFIED,
}

INVALID_CHECKLIST_MESSAGE_TEMPLATE = """\
tool call failed. 
when verification passed, following check items must be present in checklist.\
(Note that the check item names are connected with hyphens `-`, not underscores `_`.): 
{}.
If you believe these check items have already been verified, please pass them in. \
If you forgot to verify them, you may continue with your verification phase \
and call this tool after the verification has been completed successfully.
"""


class Verifier(QAAgent, ReactAgent):
    def __init__(self):
        super().__init__()
        self._mandatory_checklist: dict[str, str] = DEFAULT_VERIFIER_MANDATORY_CHECKLIST
        self._verification_result_status: VerificationResultStatus | None = None
        self._prev_checklist_arg_invalid_count = 0

    def _react_exit_prompt(self) -> str:
        return VERIFIER_FINISH_INSTRUCTION

    async def _on_turn_done(self, ctx: MessageContext) -> None:
        if self._verification_result_status is None:
            self._step_count += 1
            if self._step_count < VERIFIER_MAX_STEPS:
                await self._do_turn(ctx)
            else:
                await self._debug_log(f"exceeded max steps: {VERIFIER_MAX_STEPS}")
                self._verification_result_status = (
                    VerificationResultStatus.STEPS_EXHAUSTED
                )
                self._messages.append(
                    {
                        "role": "user",
                        "content": VERIFIER_STEPS_EXHAUSTED_PROMPT,
                    }
                )
                report_content = await self._gen_report(ctx)
                await self._send_report(report_content, ctx)
        else:
            # if self._tool_message.content:
            #     report_content = self._tool_message.content
            # else:
            report_content = await self._gen_report(ctx)

            await self._send_report(report_content, ctx)

    async def finish_verification(
        self,
        verification_result_status: VerificationResultStatus,
        checklist: list[str] | None = None,
    ) -> str:
        """结束验证并给出验证结果。"""
        if verification_result_status == VerificationResultStatus.PASSED:
            if checklist is None:
                checklist = []
            missing_checklist_items = []
            for check_item in self._mandatory_checklist.keys():
                if check_item not in checklist:
                    missing_checklist_items.append(check_item)
            if len(missing_checklist_items) > 0:
                self._prev_checklist_arg_invalid_count += 1
                if self._prev_checklist_arg_invalid_count <= 3:
                    raise RuntimeError(
                        INVALID_CHECKLIST_MESSAGE_TEMPLATE.format(
                            missing_checklist_items
                        )
                    )
                else:
                    await self._debug_log(
                        f"The LLM failed to correct the checklist parameter in the finish_verification call. To prevent an infinite loop, the verification step was allowed to proceed. {missing_checklist_items}"
                    )
        self._verification_result_status = verification_result_status
        self._prev_checklist_arg_invalid_count = 0
        return TOOL_RESULT_SUCCESS_CONTENT + ". " + VERIFIER_REPORT_INSTRUCTION

    def _get_finish_verification_tool_description(self) -> str:
        description_segs = ["结束验证并给出验证结果。"]
        if self._mandatory_checklist:
            description_segs.append(
                "在给出验证通过的结论前，必须完成以下check items的检查和验证，如果验证没有问题，则将这些check items添加到checklist参数中。"
            )
            for (
                check_item,
                check_item_desc,
            ) in self._mandatory_checklist.items():
                description_segs.append(
                    f"## check_item: `{check_item}`\n{check_item_desc}"
                )
        description = "\n".join(description_segs)
        return description

    async def _init_verification(self, task: Task) -> None:
        self._verification_result_status = None

        self._internal_tool_registry = ToolRegistry(
            [
                Tool.from_callable(
                    self.finish_verification,
                    FINISH_VERIFICATION_TOOL_NAME,
                    self._get_finish_verification_tool_description(),
                ),
            ],
        )
        self._tools = [
            tool_def.to_openai_format()
            for tool_def in chain(
                self._model_config.tools_definition,
                self._internal_tool_registry.get_tools_definition(),
            )
        ]

        verifier_instruction = VERIFIER_MAIN_INSTRUCTION
        if task.env_notes:
            verifier_instruction += task.env_notes

        personal_dir = self._id.name
        await self._publish_output_event(CreateAgentFolder(folder_name=personal_dir))

        prompt = "\n".join(
            [
                task.to_prompt(),
                self._task_config.to_prompt(personal_dir),
            ]
        )

        self._messages = [
            {"role": "system", "content": verifier_instruction},
            {"role": "user", "content": prompt},
        ]

        await self._publish_model_input(self._messages)

    @message_handler()
    async def start_verification(
        self, verification_request: VerificationRequest, ctx: MessageContext
    ) -> None:
        if len(self._messages) == 0:
            if verification_request.task.verifier_mandatory_checklist:
                self._mandatory_checklist.update(
                    verification_request.task.verifier_mandatory_checklist
                )
            await self._init_verification(verification_request.task)
        self._step_count = 0
        self._verification_result_status = None

        self._messages.append(
            {
                "role": "user",
                "content": f"planner认为当前任务已经完成，现在你需要进行验证。planner给出的task_summary如下: <task_summary>{verification_request.task_execution_summary}</task_summary>",
            }
        )

        await self._do_turn(ctx)

    async def _gen_report(self, ctx: MessageContext) -> str:
        choice = await self._acompletion_without_tool_calls(
            messages=self._messages,
            tools=self._tools,
            cache_control_injection_points=[
                {
                    "location": "message",
                    "index": -3,
                }
            ],
        )
        self._messages.append(choice.message)
        assert choice.message.content
        return choice.message.content

    async def _send_report(self, report_content: str, ctx: MessageContext):
        from judy_core.agents.planner import Planner

        assert self._verification_result_status is not None

        await ctx.send(
            VerificationResult(
                member_id=TeamMemberId.from_actor_id(self._id),
                status=self._verification_result_status,
                report_content=report_content,
            ),
            ActorId.default(Planner.actor_kind),
        )



from judy_core.agents.configuration import Configuration
from judy_core.agents.notification_center import NotificationCenter
from tongcell.core import ActorId

CONFIGURATION_ACTOR_ID = ActorId.default(Configuration.actor_kind)
NOTIFICATION_CENTER_ACTOR_ID = ActorId.default(NotificationCenter.actor_kind)

VERIFIER_NAME = "verifier-0"


def get_agent_name(prefix: str, suffix: str) -> str:
    return "{}-{}".format(prefix, suffix)


def get_executor_name(suffix: str):
    return get_agent_name("executor", suffix)
